# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Callbacks classes
"""


class Callback(object):
    """ Base class for callbacks.

    The hook_name is the path to the hook, it could be either a module, like
    "package.module", or a specific class "package.module::Class".
    The hook_name is the name of the function to hook_on, it's relative to
    the hook_module, for example with a hook_module equal to
    "package.module::Class" and a hook_path equal to "method", we will
    hook on the method named "method" of a class named "Class" in the module
    named "package.module"
    """

    def __init__(self, hook_module, hook_name, strategy=None):
        self.hook_module = hook_module
        self.hook_name = hook_name
        self.strategy = strategy

    def exception_infos(self, infos={}):
        """ Returns infos in case of exception
        """
        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Record request context
"""
from logging import getLogger
from cgi import escape

from .regexp_rule import RegexpRule
from ..runtime_infos import runtime
from ..utils import is_string

LOGGER = getLogger(__name__)


class ReflectedXSSCB(RegexpRule):
    def post(self, original, _return, *args, **kwargs):
        """ Check if a template node returns a content that is in the
        query parameters
        """
        request = runtime.get_current_request()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        if not is_string(_return):
            LOGGER.debug('Non string passed, type %s', type(_return))
            return

        if request.params_contains(_return):
            # If the payload is malicious, record the attack
            if self.match_regexp(_return):
                self.record_attack({'found': _return})

            # Only if the callback should block, sanitize the string
            if self.should_block():
                return {'status': 'override', 'new_return_value': self._escape(_return)}

    def _escape(self, value):
        """ Escape a malicious value to make it safe
        """
        return escape(value, True)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Contains all callback classes
"""
import sys
import logging

from ..rules import RuleCallback
from ..remote_exception import RemoteException

# Import rules callbacks
from .record_request_context import RecordRequestContext
from .record_request_context_django import RecordRequestContextDjango
from .record_request_context_flask import RecordRequestContextFlask
from .regexp_rule import RegexpRule
from .headers_insert import HeadersInsertCB
from .headers_insert_flask import HeadersInsertCBFlask
from .user_agent_matches import UserAgentMatchesCB
from .user_agent_matches_framework import UserAgentMatchesCBDjango, UserAgentMatchesCBFramework
from .reflected_xss import ReflectedXSSCB
from .not_found import NotFoundCB
from .not_found_django import NotFoundCBDjango
from .not_found_flask import NotFoundCBFlask
from .count_http_codes import CountHTTPCodesCB
from .count_http_codes_flask import CountHTTPCodesCBFlask
from .auth_metrics import AuthMetricsCB
from .crawler_user_agent_matches_metrics import CrawlerUserAgentMatchesMetricsCB

try:
    from .js import JSCB
except ImportError:
    pass


LOGGER = logging.getLogger(__name__)


def cb_from_rule(rule, runner):
    """ Instantiate the right cb class
    """
    # Fixme clean

    if not isinstance(rule, dict):
        LOGGER.debug("Invalid rule type %s", type(rule))
        return

    callback_class_name = rule.get('hookpoint', {}).get('callback_class')

    if callback_class_name is None:
        err_msg = "Couldn't find a callback_class_name for rule %s, fallback on js"
        LOGGER.debug(err_msg, rule['name'])
        callback_class_name = "JSCB"

    possible_subclass = globals().get(callback_class_name, None)

    if possible_subclass and issubclass(possible_subclass, RuleCallback):
        try:
            return possible_subclass.from_rule_dict(rule, runner)
        except Exception:
            LOGGER.warning("Couldn't instantiate a callback for rule %s", rule,
                           exc_info=True)
            infos = {'rule_name': rule['name'],
                     'rulespack_id': rule['rulespack_id']}
            remote_exception = RemoteException(sys.exc_info(), infos)
            runner.queue.put(remote_exception)
            return

    LOGGER.debug("Couldn't find the class matching class_name %s", callback_class_name)
    return


__all__ = ['RecordRequestContext', 'RecordRequestContextDjango',
           'RecordRequestContextFlask', 'RegexpRule', 'HeadersInsertCB',
           'UserAgentMatchesCB', 'HeadersInsertCBFlask', 'UserAgentMatchesCBDjango',
           'UserAgentMatchesCBFramework', 'ReflectedXSSCB',
           'NotFoundCB', 'NotFoundCBDjango', 'NotFoundCBFlask', 'CountHTTPCodesCB',
           'CountHTTPCodesCBFlask', 'AuthMetricsCB',
           'CrawlerUserAgentMatchesMetricsCB', 'JSCB', 'cb_from_rule']



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Generic WSGI HTTP Request / Response stuff
"""
from logging import getLogger
from itertools import chain
from copy import copy
from traceback import format_stack

from .base import BaseRequest

try:
    from Cookie import SimpleCookie
except ImportError:
    from http.cookies import SimpleCookie

try:
    from urllib.parse import parse_qs
except ImportError:
    from urlparse import parse_qs


LOGGER = getLogger(__name__)


class WSGIRequest(BaseRequest):
    """ Helper around raw wsgi environ
    """

    def __init__(self, environ):
        self.environ = environ

        form_environ = copy(self.environ)
        form_environ['QUERY_STRING'] = ''

        # TODO: Reactivate reading the body

        self.cookies = self._parse_cookies(form_environ.get('HTTP_COOKIES'))

    @staticmethod
    def _parse_cookies(cookies):
        if not cookies:
            return {}

        cookie = SimpleCookie()
        cookie.load(cookies)

        return {key: cookie[key].coded_value for key in cookie.keys()}

    @property
    def query_params(self):
        """ Return parsed query string from request
        """
        raw_query = self.environ.get('QUERY_STRING', '')
        try:
            return parse_qs(raw_query)
        except Exception:
            LOGGER.warning("Exception while parsing %s", raw_query, exc_info=True)
            return {}

    @property
    def form_params(self):
        # TODO: Reactivate reading the body
        return {}

    @property
    def cookies_params(self):
        return self.cookies

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        return list(chain.from_iterable(self.query_params.values()))

    @property
    def client_ip(self):
        if 'HTTP_X_FORWARDED_FOR' in self.environ:
            return self.environ['HTTP_X_FORWARDED_FOR'].split(",")[0].strip()
        return self.environ.get('REMOTE_ADDR')

    @property
    def hostname(self):
        return self.environ.get('HTTP_HOST', self.environ.get('SERVER_NAME'))

    @property
    def method(self):
        return self.environ.get('REQUEST_METHOD')

    @property
    def client_user_agent(self):
        return self.environ.get('HTTP_USER_AGENT')

    @property
    def referer(self):
        return self.environ.get('HTTP_REFERER')

    @property
    def scheme(self):
        return self.environ.get('wsgi.url_scheme')

    @property
    def server_port(self):
        return self.environ.get('SERVER_PORT')

    @property
    def remote_port(self):
        return self.environ.get('REMOTE_PORT')

    @property
    def path(self):
        return self.environ.get('PATH_INFO')

    def get_header(self, name):
        """ Get a specific header name
        """
        return self.environ.get(name)

    @property
    def caller(self):
        return format_stack()

    @property
    def view_params(self):
        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Instrumentation helpers
"""
import threading
from functools import wraps

LOCAL = threading.local()


def guard_call(key, callback, *args, **kwargs):
    """ Conditionnaly call callback checking that it cannot be called
    recursively
    """
    # Set guard_call_set if not present, required for multi-threaded env
    if not hasattr(LOCAL, 'guard_call_set'):
        LOCAL.guard_call_set = set()

    if key in LOCAL.guard_call_set:
        return {'status': 'not_executed'}

    LOCAL.guard_call_set.add(key)

    try:
        output = callback(*args, **kwargs)
        return output
    finally:
        try:
            LOCAL.guard_call_set.remove(key)
        except KeyError:
            pass


def partial(func, *base_args):
    """ A replacement for functools.partial that works
    with instances.
    """

    @wraps(func)
    def newfunc(*args, **kwargs):

        final_args = base_args + args

        return func(*final_args, **kwargs)

    return newfunc


class Proxy(object):
    """ An object proxy helper, forward everything to the proxied object
    except attributes starting with _sqreen
    Comes from http://code.activestate.com/recipes/496741-object-proxying/
    Licensed under PSF (https://en.wikipedia.org/wiki/Python_Software_Foundation_License)
    """

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # proxying (special cases)
    #
    def __getattr__(self, name):
        if name.startswith('_sqreen'):
            return getattr(self, name)
        else:
            return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        if name.startswith('_sqreen'):
            object.__setattr__(self, name, value)
        else:
            setattr(object.__getattribute__(self, "_obj"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    #
    # factories
    #
    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__',
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__',
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__',
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__',
        '__truediv__', '__xor__', 'next',
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)
            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Look for 404 error
"""
import logging

from ..rules import RuleCallback
from ..runtime_infos import runtime

LOGGER = logging.getLogger(__name__)


def wraps_start_response(original_start_response, callback):
    """ Decorator for start_response
    """

    def custom_start_response(status, response_headers, exc_info=None):
        """ Actually call to start_response, if status is 404 record an attack
        """

        if '404' in status:
            request = runtime.get_current_request()
            if request is not None:
                infos = {
                    'path': request.path,
                    'host': request.hostname,
                    'verb': request.method,
                    'ua': request.client_user_agent
                }
                callback.record_attack(infos)

        return original_start_response(status, response_headers, exc_info)

    return custom_start_response


class NotFoundCB(RuleCallback):
    def pre(self, original, *args):
        """ Decorate the start_response parameter to append our header on call
        """
        new_args = list(args)
        new_args[-1] = wraps_start_response(args[-1], self)
        return {'status': 'modify_args', 'args': [new_args, {}]}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Execute and sanitize remote commands
"""

import logging

from .rules_callbacks import cb_from_rule


LOGGER = logging.getLogger(__name__)


class Command(object):
    """ Abstract command
    """
    pass


class InvalidCommand(Exception):
    """ An exception raised if a command is invalid.
    It can either doesn't contains a name attribute or no callback
    has been registered for this command.
    """
    pass


class RemoteCommand(object):
    """ Class responsible for dispatching and executing remote commands
    """

    def __init__(self):
        self.commands = {}

    def process_list(self, commands, *args, **kwargs):
        """ Process a list of command and assemble the result
        """
        res = {}

        if commands is None:
            return res

        if isinstance(commands, dict):
            LOGGER.debug("Wrong commands type %s: %r", type(commands), commands)
            return res

        for command in commands:
            try:
                command_uuid = command['uuid']
            except KeyError:
                LOGGER.debug("Command without uuid: %s", command)
                continue

            res[command_uuid] = self.process(command, *args, **kwargs)
        return res

    def process(self, command, *args, **kwargs):
        """ Process a single command

        Check that the command is registered first then try
        to call it with *args and **kwargs parameters
        """
        # Validate the command
        try:
            self._validate_command(command)
        except InvalidCommand as exception:
            return {'status': False, 'msg': str(exception)}

        LOGGER.debug("Processing command %s", command['name'])

        # Then execute the command
        result = self.commands[command['name']](*args, **kwargs)
        return self._format_result(result)

    def _validate_command(self, command):
        """ Check if the command name has been registered
        """
        command_name = command.get('name')
        if command_name not in self.commands:
            msg = "unknown command name '{}'".format(command_name)
            raise InvalidCommand(msg)

    @staticmethod
    def _format_result(result):
        """ Format the command result for the backend
        """
        if result is None:
            return {'status': False, 'output': 'None returned'}
        elif result is True:
            return {'status': True}
        else:
            return {'status': True, 'output': result}

    def register_command(self, command_name, command):
        """ Register a command callback for command name
        """
        self.commands[command_name] = command

    @classmethod
    def with_production_commands(cls):
        """ Returns a RemoteCommand with all production commands
        already registered
        """
        remote_command = cls()

        for command_name, command in ALL_COMMANDS.items():
            remote_command.register_command(command_name, command)

        return remote_command


###
# COMMANDS DEFINITION
###

def _load_rules(runner):
    """ Retrieve the rulespack and instantiate the callbacks, returns
    a list of callbacks
    """
    rules = runner.session.get_rulespack()

    if len(rules['rules']) == 0 or rules['pack_id'] is None:
        return None, []

    LOGGER.info("Retrieved rulespack id: %s", rules['pack_id'])
    rules_name = ", ".join(rule['name'] for rule in rules['rules'])
    LOGGER.debug("Retrieved %d rules: %s", len(rules['rules']), rules_name)

    # Create the callbacks
    callbacks = []
    for rule_dict in rules['rules']:
        # Set the pack_id on each rule
        rule_dict['rulespack_id'] = rules['pack_id']

        # Instantiate the rule callback
        callback = cb_from_rule(rule_dict, runner)

        if callback:
            LOGGER.debug('Rule "%s" will hook on "%s %s" with callback %s with strategy %s',
                         callback.rule_name, callback.hook_module,
                         callback.hook_name, callback, callback.strategy)

            callbacks.append(callback)

            # Check if the rule has some metrics to register
            for metric in rule_dict.get('metrics', []):
                runner.metrics_store.register_metric(**metric)

    return rules['pack_id'], callbacks


def _instrument_callbacks(runner, callbacks):
    """ For a given list of callbacks, hook them
    """
    LOGGER.info("Setup instrumentation")
    for callback in callbacks:
        runner.instrumentation.add_callback(callback)


def instrumentation_enable(runner):
    """ Retrieve a rulespack, instantiate RuleCallback from the rules
    and instrument them.
    """
    # Load rules
    pack_id, rules = _load_rules(runner)

    LOGGER.debug("Start instrumentation with rulespack_id %s", pack_id)

    # Clean existing callbacks to avoid double-instrumentation
    instrumentation_remove(runner)

    # Instrument retrieved rules
    _instrument_callbacks(runner, rules)

    return pack_id


def instrumentation_remove(runner):
    """ Remove all callbacks from instrumentation, return True
    """
    LOGGER.debug("Remove instrumentation")
    runner.instrumentation.deinstrument_all()
    return True


def rules_reload(runner):
    """ Load the rules, deinstrument the old ones and instrument
    the new loaded rules. Returns the new rulespack_id
    """
    LOGGER.debug("Reloading rules")
    pack_id, rules = _load_rules(runner)

    # Deinstrument
    instrumentation_remove(runner)

    # Reinstrument new ones
    _instrument_callbacks(runner, rules)

    LOGGER.debug('Rules reloaded')
    return pack_id


def features_get(runner):
    print("COMMAND: Features_get")


def features_change(runner):
    print("COMMAND: Features change")

ALL_COMMANDS = {
    "instrumentation_enable": instrumentation_enable,
    "instrumentation_remove": instrumentation_remove,
    "rules_reload": rules_reload,
    "features_get": features_get,
    "features_change": features_change
}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Sqreen attack event helpers and placeholder
"""
import traceback
from logging import getLogger

try:
    from http.cookies import BaseCookie
except ImportError:
    from Cookie import BaseCookie

try:
    from urllib.parse import parse_qs
except ImportError:
    from urlparse import parse_qs

from .remote_exception import traceback_formatter


LOGGER = getLogger(__name__)


def get_context_payload():
    """ Return attack payload dependent on the context, right now stacktrace.
    """
    return {
        'context': {
            'backtrace': list(traceback_formatter(traceback.extract_stack()))
        }
    }


class Attack(object):

    def __init__(self, payload):
        self.payload = payload
        self.rule_name = payload['rule_name']

    def to_dict(self):
        return self.payload



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Insert X-Protected-By header
"""
import logging
from collections import Mapping
from copy import copy

from ..rules import RuleCallback
from ..exceptions import InvalidArgument


LOGGER = logging.getLogger(__name__)


def wraps_start_response(original_start_response, callback):
    """ Decorator for start_response
    """

    def custom_start_response(status, response_headers, exc_info=None):
        """ Actually call to start_response, append our header
        """
        callback_headers = copy(callback.headers)

        # Override existing headers
        for i, header in enumerate(response_headers):
            header_name = header[0]
            if header_name in callback_headers.keys():
                response_headers[i] = (header_name, callback_headers.pop(header_name))

        # Add the remaining custom headers
        response_headers += list(callback_headers.items())

        return original_start_response(status, response_headers, exc_info)

    return custom_start_response


def convert_to_str(headers):
    """ Encode a list of headers tuples into latin1
    """
    for header_name, header_value in headers:
        header_name = str(header_name.encode('latin-1', errors='replace').decode('latin-1'))
        header_value = str(header_value.encode('latin-1', errors='replace').decode('latin-1'))
        yield (header_name, header_value)


class HeadersInsertCB(RuleCallback):
    """ Callback that add the custom sqreen header
    """

    def __init__(self, *args, **kwargs):
        super(HeadersInsertCB, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            self.values = self.data['values']
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        self._headers = None

    @property
    def headers(self):
        """ Cached property to defer headers data conversio,n
        """
        if self._headers is None:
            self._headers = dict(convert_to_str(self.values))
        return self._headers

    def pre(self, original, *args):
        """ Decorate the start_response parameter to append our header on call
        """
        new_args = list(args)
        new_args[-1] = wraps_start_response(args[-1], self)
        return {'status': 'modify_args', 'args': [new_args, {}]}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Count http codes for flask framework
"""
import logging

from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class CountHTTPCodesCBFlask(RuleCallback):

    def post(self, original, response, *args, **kwargs):
        """ Recover the status code and update the http_code metric
        """
        status_code = response.status_code
        self.record_observation('http_code', str(status_code), 1)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Count http codes
"""
import logging

from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


def wraps_start_response(original_start_response, callback):
    """ Decorator for start_response
    """

    def custom_start_response(status, response_headers, exc_info=None):
        """ Actually call to start_response, if status is 404 record an attack
        """

        try:
            if status:
                status_code = None
                if isinstance(status, str):
                    status_code = status.split(' ', 1)[0]
                elif isinstance(status, bytes):
                    status_code = status.split(b' ', 1)[0].decode('latin_1')

                if status_code:
                    callback.record_observation('http_code', status_code, 1)
        except Exception:
            LOGGER.exception("Exception in %s", callable.__class__.__name__)

        return original_start_response(status, response_headers, exc_info)

    return custom_start_response


class CountHTTPCodesCB(RuleCallback):
    def pre(self, original, *args):
        """ Decorate the start_response parameter to capture http status code
        """
        new_args = list(args)
        new_args[-1] = wraps_start_response(args[-1], self)
        return {'status': 'modify_args', 'args': [new_args, {}]}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Record request context
"""

from ..rules import RuleCallback
from ..runtime_infos import runtime
from ..frameworks.wsgi import WSGIRequest


class RecordRequestContext(RuleCallback):

    @staticmethod
    def pre(*args, **kwargs):
        runtime.store_request(WSGIRequest(args[-2]))

    @staticmethod
    def post(*args, **kwargs):
        runtime.clear_request()

    @staticmethod
    def failing(*args, **kwargs):
        runtime.clear_request()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Blank request for callbacks needing a request when no one is present
"""

from .base import BaseRequest


class BlankRequest(BaseRequest):

    @property
    def client_ip(self):
        return None

    @property
    def client_user_agent(self):
        return None

    @property
    def cookies_params(self):
        return {}

    @property
    def form_params(self):
        return {}

    @property
    def hostname(self):
        return None

    @property
    def method(self):
        return None

    @property
    def path(self):
        return None

    @property
    def query_params(self):
        return {}

    @property
    def query_params_values(self):
        return []

    @property
    def referer(self):
        return None

    @property
    def remote_port(self):
        return None

    @property
    def scheme(self):
        return None

    @property
    def server_port(self):
        return None

    @property
    def view_params(self):
        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Look for badly behaved clients
"""
import sys
import logging
from .regexp_rule import RegexpRule

LOGGER = logging.getLogger(__name__)

ERROR_STRING = ("<html><head><title>Internal Server Error</title></head>"
                "<body><h1><p>Internal Server Error</p></h1></body></html>")


class UserAgentMatchesCB(RegexpRule):

    def pre(self, original, *args, **kwargs):
        environ = args[-2]
        start_response = args[-1]

        user_agent = environ.get('HTTP_USER_AGENT')
        if not user_agent:
            return

        match = self.match_regexp(user_agent)
        if not match:
            return

        infos = {'found': match}
        self.record_attack(infos)

        # Try to detect gunicorn
        server_is_gunicorn = environ.get('SERVER_SOFTWARE', '').startswith('gunicorn')
        is_gevent_loaded = 'gevent' in sys.modules
        if server_is_gunicorn and is_gevent_loaded and self.should_block():
            # Log the error
            err = "Sqreen blocked a security threat (type: #%s). No action is required."
            LOGGER.error(err, self.rule_name)

            # Return a 500 error instead for gunicorn
            start_response('500 Internal Server Error',
                           [('Content-Type', 'text/html'),
                            ('Content-Length', str(len(ERROR_STRING)))])

            return {'status': 'override', 'new_return_value': [ERROR_STRING]}
        else:
            # Blocking is checked by the caller
            return {'status': 'raise', 'data': match,
                    'rule_name': self.rule_name}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Django hook strategy
"""

import sys

from logging import getLogger

from .base import BaseStrategy
from ..hook_point import DjangoMiddleware
from ..import_hook import ImportHook, get_hook_parent


LOGGER = getLogger(__name__)


def load_middleware_insert(original, middleware):

    def wrapped_load_middleware(self, *args, **kwargs):
        LOGGER.debug("Execute load_middleware_insert")

        # Load original middlewares
        result = original(self, *args, **kwargs)

        # Insert out custom one
        try:
            self._view_middleware.insert(0, middleware.process_view)
            self._response_middleware.append(middleware.process_response)
            self._exception_middleware.append(middleware.process_exception)
        except Exception:
            LOGGER.warning("Error while inserting our middleware", exc_info=True)

        return result

    return wrapped_load_middleware


class DjangoStrategy(BaseStrategy):
    """ Strategy for Django peripheric callbacks.

    It injects a custom DjangoFramework that calls callbacks for each
    lifecycle method
    """

    MODULE_NAME = "django.core.handlers.base"
    HOOK_CLASS = "BaseHandler"
    HOOK_METHOD = "load_middleware"

    def __init__(self, strategy_key, channel, before_hook_point=None):
        super(DjangoStrategy, self).__init__(channel, before_hook_point)
        self.strategy_key = strategy_key

        self.django_middleware = DjangoMiddleware(self, channel)

        self.hooked = False

    def hook(self, callback):
        """ Accept a callback and store it. If it's the first callback
        for this strategy, actually hook the load_middleware to insert our
        middleware.

        Once hooked, the middleware will call the callbacks at the right moment.
        """
        # Register callback
        self.add_callback(callback)

        # Check if we already hooked at
        if not self.hooked:

            import_hook = ImportHook(self.MODULE_NAME, self.import_hook_callback)
            sys.meta_path.insert(0, import_hook)

            self.hooked = True

    def import_hook_callback(self, module):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module
        """
        hook_parent = get_hook_parent(module, self.HOOK_CLASS)

        original = getattr(hook_parent, self.HOOK_METHOD, None)
        hooked = load_middleware_insert(original, self.django_middleware)
        setattr(hook_parent, self.HOOK_METHOD, hooked)
        LOGGER.debug("Successfully hooked on %s %s", self.MODULE_NAME,
                     self.HOOK_CLASS)

    @classmethod
    def get_strategy_id(cls, callback):
        """ This strategy only hook on
        (django.core.handlers.base::BaseHandler, load_middleware)
        """
        return ("{}::{}".format(cls.MODULE_NAME, cls.HOOK_CLASS), cls.HOOK_METHOD)

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Helpers to retrieve runtime informations
"""
import os
import sys
import datetime
import platform
import threading

from logging import getLogger

try:
    from pip import get_installed_distributions
    HAS_PIP = True
except ImportError:
    # We don't have pip, fallback on pkg_resources
    HAS_PIP = False
    import pkg_resources

    def get_installed_distributions():
        """ Mimic pip.get_installed_distributions using pkg_resources
        """
        return pkg_resources.working_set


import sqreen

LOGGER = getLogger(__name__)


def _format_datetime_isoformat(datetime_object):
    """ Take a datetime and return it as isoformat
    """
    return datetime_object.isoformat()


class RuntimeStorage(object):
    """ Object for storing runtime informations like the current request processed
    """

    def __init__(self):
        self.local = threading.local()
        self._current_request_initializer()

    def _current_request_initializer(self):
        if not hasattr(self.local, "current_request"):
            setattr(self.local, "current_request", None)

    def store_request(self, request):
        """ Store a request, whatever its kind in a thread-safe way
        """
        LOGGER.debug("Store request %s", request)
        self._current_request_initializer()
        self.local.current_request = request

    def clear_request(self):
        """ Clear the stored request in a thread-safe way
        """
        if hasattr(self.local, "current_request"):
            LOGGER.debug("Clear request %s", getattr(self.local, 'current_request'))
            self.local.current_request = None

    def get_current_request(self):
        """ Return current processed request
        """
        self._current_request_initializer()
        return self.local.current_request


runtime = RuntimeStorage()


class RuntimeInfos(object):
    """ Helper to collect and return environement informations about the
    runtime
    """

    def all(self):
        """ Returns aggregated infos from the environment
        """
        resultat = {'various_infos': {}}
        resultat.update(self._agent())
        resultat.update(self._framework())
        resultat.update(self._os())
        resultat.update(self._runtime())
        resultat['various_infos'].update(self._time())
        resultat['various_infos'].update(self._dependencies())
        resultat['various_infos'].update(self._process())
        return resultat

    @staticmethod
    def _dependencies():
        """ Returns informations about the installed python dependencies
        """
        dep = [{
            'name': package.project_name,
            'version': package.version
        } for package in get_installed_distributions()]
        return {"dependencies": dep}

    @staticmethod
    def _time():
        """ Returns informations about the current time
        """
        return {'time': _format_datetime_isoformat(datetime.datetime.utcnow())}

    @staticmethod
    def _agent():
        """ Returns informations about the agent
        """
        return {'agent_type': 'python', 'agent_version': sqreen.__version__}

    @staticmethod
    def _get_package_version(package_name):
        for package in get_installed_distributions():
            if package.project_name == package_name:
                return package.version

    def _framework(self):
        """ Returns informations about the web framework
        """
        flask = self._get_package_version('Flask')
        django = self._get_package_version('Django')
        pyramid = self._get_package_version('pyramid')

        if django:
            return {'framework_type': 'Django', 'framework_version': django}
        elif flask:
            return {'framework_type': 'Flask', 'framework_version': flask}
        elif pyramid:
            return {'framework_type': 'pyramid', 'framework_version': pyramid}
        else:
            return {'framework_type': None, 'framework_version': None}

    @staticmethod
    def _os():
        """ Returns informations about the os
        """

        # Compute the OS version
        if sys.platform == 'darwin':
            base_os_version = 'Mac OS X {}'.format(platform.mac_ver()[0])
        elif 'linux' in sys.platform:
            base_os_version = '{0[0]} {0[1]}'.format(platform.linux_distribution())
        else:
            base_os_version = ''

        os_version = '{}/{}'.format(base_os_version, '/'.join(os.uname()))
        return {'os_type': '{}-{}'.format(platform.machine(), sys.platform),
                'os_version': os_version,
                'hostname': platform.node()}

    @staticmethod
    def _runtime():
        """ Returns informations about the python runtime
        """
        python_build = platform.python_build()
        version = '{} ({}, {})'.format(platform.python_version(), python_build[0],
                                       python_build[1])
        return {'runtime_type': platform.python_implementation(),
                'runtime_version': version}

    @staticmethod
    def _process():
        """ Returns informations about the current process
        """
        return {
            'pid': os.getpid(),
            'ppid': os.getppid(),
            'euid': os.geteuid(),
            'egid': os.getegid(),
            'uid': os.getuid(),
            'gid': os.getgid(),
            # sys.argv is not always available
            'name': getattr(sys, 'argv', [None])[0]
        }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Django specific WSGI HTTP Request / Response stuff
"""
from itertools import chain
from logging import getLogger

from .base import BaseRequest


LOGGER = getLogger(__name__)


class DjangoRequest(BaseRequest):

    def __init__(self, request, view_func, view_args, view_kwargs):
        self.request = request
        self.view_func = view_func
        self.view_args = view_args
        self.view_kwargs = view_kwargs

        # Convert django QueryDict to a normal dict with values as list
        self.converted_get = dict(self.request.GET.lists())

    @property
    def query_params(self):
        return self.converted_get

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        return list(chain.from_iterable(self.converted_get.values()))

    @property
    def form_params(self):
        return dict(self.request.POST)

    @property
    def cookies_params(self):
        return self.request.COOKIES

    @property
    def client_ip(self):
        if 'HTTP_X_FORWARDED_FOR' in self.request.META:
            return self.request.META['HTTP_X_FORWARDED_FOR'].split(",")[0].strip()
        return self.request.META.get('REMOTE_ADDR')

    @property
    def hostname(self):
        return self.request.get_host()

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return self.request.META.get('HTTP_REFERER')

    @property
    def client_user_agent(self):
        return self.request.META.get('HTTP_USER_AGENT')

    @property
    def path(self):
        return self.request.get_full_path()

    @property
    def scheme(self):
        return self.request.scheme

    @property
    def server_port(self):
        return self.request.META.get('SERVER_PORT')

    @property
    def remote_port(self):
        return self.request.META.get('REMOTE_PORT')

    def get_header(self, name):
        """ Get a specific header
        """
        return self.request.META.get(name)

    @property
    def view_params(self):
        return self.view_kwargs



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Binding accessor class
"""

import re
import types


class StringScanner(object):
    """ Equivalent of http://ruby-doc.org/stdlib-1.9.3/libdoc/strscan/rdoc/StringScanner.html
    """

    def __init__(self, string):
        self.string = string
        self.pos = 0
        self.match = None

    def scan(self, regex):
        """ Tries to match with pattern at the current position.
        If there’s a match, the scanner advances the “scan pointer” and returns
        the matched string. Otherwise, the scanner returns None.
        """
        match = re.compile(regex).match(self.string, self.pos)

        if match:
            self.pos += len(match.group(0))
            self.match = match
            return match

    @property
    def eos(self):
        """ Returns True if the scan pointer is at the end of the string
        """
        return self.pos == len(self.string)

# According to https://docs.python.org/2/reference/lexical_analysis.html#identifiers
PYTHON_IDENTIFIER = r"([a-zA-Z]|_)([a-zA-Z]|\d|_)*"


class BindingAccessor(object):
    """ Given a binding accessor expression (#.args[0]) when giving a context
    resolve the expression and returns the value corresponding.
    """

    def __init__(self, expression):
        self.path = []
        self.expression = expression
        self.parse(expression)

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.expression)

    def resolve(self, binding, global_binding=None, framework=None, instance=None,
                arguments=None, kwarguments=None, cbdata=None, return_value=None):
        """ Takes a context, resolve the expression and return the value
        """
        value = None
        env = [framework, instance, arguments, kwarguments, cbdata, return_value]

        if global_binding is None:
            global_binding = {}

        for component in self.path:
            value = self._resolve_component(value, component, binding, global_binding, env)

        self._validate_value(value)

        return value

    def _resolve_component(self, value, component, binding, global_binding, env):
        """ Resolve simple expression, like static string, integer, index or
        attribute. Dispatch to other function in other cases.
        """
        if component["kind"] in ('string', 'integer'):
            return component["value"]
        elif component["kind"] == "variable":
            return self._resolve_variable(component["name"], binding, global_binding)
        elif component["kind"] == "index":
            return value[component["index"]]
        elif component["kind"] == "attribute":
            return getattr(value, component["name"])
        elif component["kind"] == "sqreen-variable":
            return self._resolve_sqreen_variable(component["name"], *env)
        else:
            raise ValueError("Uknown component {}".format(component))

    @staticmethod
    def _resolve_sqreen_variable(what, framework, instance, arguments,
                                 kwarguments, cbdata, return_value):
        """ Resolve sqreen-variables (the one starting with #.).
        Fallback on the framework object if it's not a special sqreen-variable
        """
        if what == 'data':
            return cbdata
        elif what == 'rv':
            return return_value
        elif what == 'args':
            return arguments
        elif what == 'kwargs':
            return kwarguments
        elif what == 'inst':
            return instance
        else:
            # Default to None if the framework is None
            if framework is None:
                return None

            return getattr(framework, what)

    def _resolve_variable(self, variable_name, binding, global_binding):
        """ Resolve general variable name, foo, bar. Search first in local
        context then in general contetx
        """
        if variable_name in binding:
            return binding[variable_name]
        elif variable_name in global_binding:
            return global_binding[variable_name]
        else:
            raise NameError("name '{}' was not found in bindings".format(variable_name))

    def _validate_value(self, value):
        """ In case of invalid values, raise an Exception instead
        """
        invalid_types = (types.FunctionType, types.MethodType)

        if isinstance(value, invalid_types):
            raise ValueError("Invalid return: {!r}".format(value))

    ###
    # Parse methods
    ###
    def parse(self, expression):
        """ Parse a binding accessor expression and convert it to a serie
        of instruction to resolve later.
        """
        self.scanner = StringScanner(expression)
        while not self.scanner.eos:
            initial_pos = self.scanner.pos

            # Check for scalar first
            scalar = self.scan_scalar()
            if scalar:
                self.path.append(scalar)
                return

            # If we are at the beginning of an expression, a variable is allowed
            if initial_pos == 0:
                self.scan_push_variable()
            else:
                self.scan_push_attribute()

            self.scan_push_indexes()

            # Remove potential dot
            self.scanner.scan(r'\.')

            if initial_pos == self.scanner.pos:
                raise ValueError('Parsing error, parser is stuck')

        # Do not keep scanner instance around
        del self.scanner

    def scan_push_variable(self):
        if self.scanner.scan(r'#\.(\w+)'):
            self.path.append({"name": self.scanner.match.group(1), "kind": "sqreen-variable"})
        elif self.scanner.scan(PYTHON_IDENTIFIER):
            self.path.append({"name": self.scanner.match.group(), "kind": "variable"})

    def scan_push_attribute(self):
        if self.scanner.scan(PYTHON_IDENTIFIER):
            self.path.append({"name": self.scanner.match.group(), "kind": "attribute"})

    def scan_scalar(self):
        if self.scanner.scan(r'\d+'):
            return {"value": int(self.scanner.match.group()), "kind": 'integer'}
        elif self.scanner.scan(r"'((?:\\.|[^\\'])*)'"):
            return {"value": self.scanner.match.group(1), "kind": 'string'}

    def scan_push_indexes(self):
        while self.scanner.scan(r"\["):
            scalar = self.scan_scalar()
            if not scalar:
                raise Exception("Invalid index")

            if not self.scanner.scan(r"\]"):
                raise ValueError("Unfinished index")

            self.path.append({"index": scalar["value"], "kind": "index"})



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Various utils
"""
import sys

from functools import WRAPPER_ASSIGNMENTS


def is_string(value):
    """ Check if a value is a valid string, compatible with python 2 and python 3

    >>> is_string('foo')
    True
    >>> is_string(u'✌')
    True
    >>> is_string(42)
    False
    >>> is_string(('abc',))
    False
    """
    python_version = sys.version_info[0]

    if python_version == 2:
        return isinstance(value, basestring)
    elif python_version == 3:
        return isinstance(value, str)
    else:
        raise NotImplementedError()


def is_unicode(value):
    """ Check if a value is a valid unicode string, compatible with python 2 and python 3

    >>> is_unicode(u'foo')
    True
    >>> is_unicode(u'✌')
    True
    >>> is_unicode(b'foo')
    False
    >>> is_unicode(42)
    False
    >>> is_unicode(('abc',))
    False
    """
    python_version = sys.version_info[0]

    if python_version == 2:
        return isinstance(value, unicode)
    elif python_version == 3:
        return isinstance(value, str)
    else:
        raise NotImplementedError()


def to_latin_1(value):
    """ Return the input string encoded in latin1 with replace mode for errors
    """
    return value.encode('latin-1', 'replace')


def update_wrapper(wrapper, wrapped):
    """ Update wrapper attribute to make it look like wrapped function.
    Don't use original update_wrapper because it can breaks if wrapped don't
    have all attributes.
    """
    for attr in WRAPPER_ASSIGNMENTS:
        if hasattr(wrapped, attr):
            setattr(wrapper, attr, getattr(wrapped, attr))
    return wrapper



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Callbacks rules classes and helpers
"""
import logging

from datetime import datetime
from functools import wraps

from .callbacks import Callback
from .events import Attack, get_context_payload
from .runtime_infos import RuntimeInfos, runtime
from .runner import MetricsEvent
from .condition_evaluator import ConditionEvaluator, is_condition_empty
from .constants import LIFECYCLE_METHODS


LOGGER = logging.getLogger(__name__)


def check_condition_wrapper(rule_data, function, condition, lifecycle):
    """ Wrapper that will check lifecycle method pre-condition before
    calling it.
    If pre-conditions are true, call the lifecycle method, otherwise
    return None.
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        """ Wrapper around lifecycle method
        """

        # Copy args to work on them
        binding_args = list(args)

        # Compute return value depending on the lifecycle method we wrap
        return_value = False
        if lifecycle in ('post', 'failing'):
            # Args[0] is either the return value of the hooked point
            # or the exception
            return_value = binding_args.pop(0)

        original = binding_args.pop(0)

        binding_eval_args = {
            "binding": locals(),
            "global_binding": globals(),
            "framework": runtime.get_current_request(),
            "instance": original,
            "arguments": binding_args,
            "cbdata": rule_data,
            "return_value": return_value
        }

        # Check the pre condition
        if condition.evaluate(**binding_eval_args) in (False, None):
            return None

        # Execute the hook otherwise with the original args
        return function(*args, **kwargs)

    wrapped.__wrapped__ = function

    return wrapped


class RuleCallback(Callback):

    def __init__(self, hook_module, hook_name, rule_name, rulespack_id,
                 block, test, runner, strategy_cls=None, data=None,
                 conditions=None, callbacks=None):
        super(RuleCallback, self).__init__(hook_module, hook_name, strategy_cls)
        self.rule_name = rule_name
        self.rulespack_id = rulespack_id

        if data is None:
            data = {}
        self.data = data
        self.block = block
        self.test = test
        self.runner = runner

        if conditions is None:
            conditions = {}
        self.conditions = conditions

        self.callbacks = callbacks

        self._apply_conditions()

    @classmethod
    def from_rule_dict(cls, rule_dict, runner):
        """ Return a RuleCallback based on a rule dict
        """
        return cls(
            hook_module=rule_dict['hookpoint']['klass'],
            hook_name=rule_dict['hookpoint']['method'],
            rule_name=rule_dict['name'],
            rulespack_id=rule_dict['rulespack_id'],
            strategy_cls=rule_dict['hookpoint'].get('strategy', 'import_hook'),
            block=rule_dict['block'],
            test=rule_dict['test'],
            data=rule_dict.get('data'),
            conditions=rule_dict.get('conditions'),
            callbacks=rule_dict.get('callbacks'),
            runner=runner
        )

    def _apply_conditions(self):
        """ Wrap each lifecycle methods if the Rule define them and if we have
        conditions for them.
        """
        for lifecycle in LIFECYCLE_METHODS.values():
            conditions = self.conditions.get(lifecycle)

            if not is_condition_empty(conditions) and hasattr(self, lifecycle):
                conditions = ConditionEvaluator(conditions)

                # Wrap the lifecycle method
                lifecycle_method = getattr(self, lifecycle)
                wrapped = check_condition_wrapper(self.data, lifecycle_method,
                                                  conditions, lifecycle)
                setattr(self, lifecycle, wrapped)

    def exception_infos(self, infos={}):
        """ Returns additional infos in case of exception
        """
        return {
            'rule_name': self.rule_name,
            'rulespack_id': self.rulespack_id,
            'should_block': self.should_block()
        }

    def should_block(self):
        """ Return True if this rule should catch an attack based on block
        and test value
        """
        return ((not self.test) and self.block)

    def record_attack(self, infos=None):
        """ Record an attack
        """
        current_request = runtime.local.current_request
        current_time = RuntimeInfos()._time()['time']

        rule_payload = {
            'rule_name': self.rule_name,
            'rulespack_id': self.rulespack_id,
            'client_ip': current_request.client_ip,
            'time': current_time,
            'test': self.test,
            'block': self.block,

            'infos': infos
        }
        rule_payload.update(current_request.full_payload)
        rule_payload.update(get_context_payload())
        attack = Attack(rule_payload)
        self.runner.queue.put(attack)

        LOGGER.debug("Attack was pushed to the queue: %s", rule_payload)

    def record_observation(self, metric_name, key, observation, at=None):
        """ Record a metric observation, push it to runner
        observations queue
        """
        if at is None:
            at = datetime.utcnow()

        self.runner.observation_queue.put((metric_name, at, key, observation))

        # Ask the runner thread to aggregate and empty the observation queue
        if self.runner.observation_queue.half_full():
            self.runner.queue.put(MetricsEvent)

    def __repr__(self):
        return "%s(rule_name=%r)" % (self.__class__.__name__, self.rule_name)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Instrumentation helper responsible for adding dynamic callback
"""
import sys
import logging

from .strategies import SetAttrStrategy, DBApi2Strategy, ImportHookStrategy
from .strategies import DjangoStrategy, Psycopg2Strategy, FlaskStrategy

from ..remote_exception import RemoteException


LOGGER = logging.getLogger(__name__)


class Instrumentation(object):
    """ The instrumentation class is the exposed face of the
    instrumentation engine. It dispatchs to the right strategy,
    the default one is set_attr.

    The instrumentation class takes a channel as parameter. The channel
    is a callable accepting 1 parameter where events, attacks and
    remote exception will be sent to. It can be CappedQueue.push or simply
    list().append

    The instrument class dispatch to different strategies based on strategy
    name defined in callback. It ask stategy for an unique id based on hook path
    infos and ensure to have only one strategy instance per id. It's needed
    for DBApi2 strategy where every sqlite3 callbacks will be stored in the same
    strategy to avoid double-instrumentation.
    """

    def __init__(self, channel, before_hook_point=None):
        self.channel = channel
        self.strategies = {}
        self.before_hook_point = before_hook_point

    def add_callback(self, callback):
        """ Add a callback. The callback defines itself where it should
        hook to and the strategy use for hooking (set_attr or DBApi2)
        """
        try:
            strategy_class = self._get_strategy_class(callback.strategy)

            # Get the strategy id
            strategy_id = strategy_class.get_strategy_id(callback)

            # Check if we already have a strategy
            if strategy_id in self.strategies:
                strategy_instance = self.strategies[strategy_id]
                LOGGER.debug("Reusing strategy %s for id %s",
                             strategy_instance, strategy_id)
            else:
                strategy_instance = strategy_class(strategy_id, self.channel,
                                                   self.before_hook_point)
                LOGGER.debug("Instantiate strategy %s for id %s -> %s",
                             strategy_class, strategy_id, strategy_instance)
                self.strategies[strategy_id] = strategy_instance

            strategy_instance.hook(callback)
        except Exception:
            # If the strategy fails to hook either at instantiation
            # or hook method, catch the exception and log it
            LOGGER.exception("Callback %r fails to hook", callback, exc_info=True)

            # And send the exception back to the backend
            infos = {'callback': callback.__dict__}
            remote_exception = RemoteException(sys.exc_info(), infos)
            self.channel(remote_exception)

    def deinstrument(self, callback):
        """ Deactive instrumentation on the callback endpoint
        """
        strategy_class = self._get_strategy_class(callback.strategy)

        # Get the strategy id
        strategy_id = strategy_class.get_strategy_id(callback)

        self.strategies[strategy_id].deinstrument(callback)

    def deinstrument_all(self):
        """ Deactive instrumentation on all callbacks by calling
        deinstrument_all on all strategies
        """
        for strategy in self.strategies.values():
            strategy.deinstrument_all()

    @staticmethod
    def _get_strategy_class(strategy):
        """ Return a strategy class depending on the strategy name passed
        in parameter.
        Raise a NotImplementedError if the strategy is unknown.
        """
        if strategy == 'set_attr':
            return SetAttrStrategy
        elif strategy == 'import_hook':
            return ImportHookStrategy
        elif strategy == 'DBApi2':
            return DBApi2Strategy
        elif strategy == 'django':
            return DjangoStrategy
        elif strategy == 'psycopg2':
            return Psycopg2Strategy
        elif strategy == 'flask':
            return FlaskStrategy
        else:
            err_msg = "Unknown hooking_strategy {}"
            raise NotImplementedError(err_msg.format(strategy))



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
from .engine import Instrumentation

__all__ = ['Instrumentation']



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Script modifying the env to run our sitecustomize script before everyone else
"""
from __future__ import print_function

import sys
import os

from copy import copy
from os.path import dirname


def where_sitecustomize():
    """ Returns the path of the sitecustomize package
    """
    current_file = os.path.split(__file__)[0]
    path = os.path.join(dirname(current_file), 'sitecustomize')

    return path


def protect():
    """ Call the passed binary passed in argument with the right PYTHONPATH
    """
    # Check that we have all mandatory arguments
    if len(sys.argv) <= 1:
        print("Usage: sqreen-start ...", file=sys.stderr)
        print("You can check https://doc.sqreen.io/v1.1/docs/python-agent-installation"
              " for more informations.", file=sys.stderr)
        sys.exit(1)

    new_env = copy(os.environ)
    if 'PYTHONPATH' in new_env:
        new_env['PYTHONPATH'] = os.pathsep.join([where_sitecustomize(), new_env['PYTHONPATH']])
    else:
        new_env['PYTHONPATH'] = where_sitecustomize()

    new_args = sys.argv[1:]
    os.execvpe(new_args[0], new_args, new_env)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Collect and aggregate metrics
"""
from datetime import timedelta

from .exceptions import SqreenException


class BaseAggregator(object):
    """ Base classes for aggregators.

    It contains the methods to update a period with new data and finalize the
    data just before uploading.
    Aggregators doesn't store any state, it is passed by the MetricsStore.
    """

    name = "Base"

    def update(self, key, value, data=None):
        """ Update the whole data for the period. For updating only the key,
        override update_key instead.
        """
        if data is None:
            data = {}

        data[key] = self.update_key(value, data.get(key))
        return data

    @staticmethod
    def update_key(value, data=None):
        """ Update the data for a key in the period. For updating the whole data,
        override update instead.
        """
        raise NotImplementedError("")

    @staticmethod
    def finalize(data):
        """ Finalize the period, by default return the data. Could be override
        to compute a ration or sum some elements.
        """
        return data


class CollectAggregator(BaseAggregator):
    """ Simple aggregator that stores all the observations points,
    used in tests mostly
    """

    name = "Collect"

    @staticmethod
    def update_key(value, data=None):
        if data is None:
            data = []

        data.append(value)
        return data


class SumAggregator(BaseAggregator):

    name = "Sum"

    @staticmethod
    def update_key(value, data=None):
        if data is None:
            data = 0

        return data + value


class AverageAggregator(BaseAggregator):

    name = "Average"

    @staticmethod
    def update_key(value, data=None):
        if data is None:
            data = {'sum': 0, 'count': 0}

        data['sum'] += value
        data['count'] += 1
        return data

    @staticmethod
    def finalize(data):
        final_data = {}

        for key, value in data.items():
            final_data[key] = value['sum'] / float(value['count'])

        return final_data


class UnknownAggregator(SqreenException):
    """ Exception raised when trying to register a metric with an unknown
    aggregation kind
    """


class AlreadyRegisteredMetric(SqreenException):
    """ Exception raised when trying to register twice the same metric
    name.
    """


class AlreadyRegisteredAggregator(SqreenException):
    """ Exception raised when trying to register twice the same aggregator
    name.
    """


class UnregisteredMetric(SqreenException):
    """ Exception raised when trying to update an unregistered metric.
    """


PRODUCTION_AGGREGATORS = [SumAggregator(), AverageAggregator()]


class MetricsStore(object):
    """ Store the dict of currently monitored metrics.

    For each metric, store a dict containing:
    - The start time of monitored period.
    - The maximum period time.
    - The kind of aggregator.
    - The aggregated data, the value is managed by the aggregator directly.

    Store also the list of available aggregator indexed by a kind.

    When periods are finished, store them to be retrieved for pushing in
    the store attribute.
    """

    def __init__(self):
        self.store = []
        self.metrics = {}
        self.aggregators = {}

    def register_metric(self, name, kind, period):
        """ Register a new metric
        """
        if name in self.metrics:
            saved_metric_kind = self.metrics[name]['kind']
            if saved_metric_kind != kind:
                err_msg = "Metric '{}' has already been registered with kind {}"
                raise AlreadyRegisteredMetric(err_msg.format(name, saved_metric_kind))

        if kind not in self.aggregators:
            raise UnknownAggregator("Unknown aggregation kind: {}".format(kind))
        self.metrics[name] = self._new_metric(kind, period)

    @staticmethod
    def _new_metric(kind, period):
        """ Return a dict for an empty metric period
        """
        return {"kind": kind, "period": period,
                "observation": None, "start": None}

    def register_aggregator(self, name, aggregator_function):
        """ Register a new aggregator under the name passed in input
        """
        if name in self.aggregators:
            msg = "Aggregator '{}' has already been registered to: {}"
            raise AlreadyRegisteredAggregator(msg.format(name, self.aggregators[name]))
        self.aggregators[name] = aggregator_function

    def register_production_aggregators(self):
        """ Register production aggregators
        """
        for aggregator in PRODUCTION_AGGREGATORS:
            self.register_aggregator(aggregator.name, aggregator)

    def update(self, name, at, key, value):
        """ Logic behind metric updating.

        Check if the start time is set for the metric period.
        Check if the metric period has expired, if so save it and create
        a blank metric period.
        Then call the aggregator to compute the new data
        """
        try:
            metric = self.metrics[name]
        except KeyError:
            raise UnregisteredMetric("Unknown metric {}".format(name))

        # Update start time if not set already (registered but never updated)
        if metric["start"] is None:
            metric["start"] = at

        # Check if the metric should be published or not
        elif metric["start"] + timedelta(seconds=metric["period"]) < at:
            metric = self.finalize_period(metric, name, at)

        self._update_metric(metric, key, value)

    def _update_metric(self, metric, key, value):
        """ Actual method that call the aggregator to compute the new data.
        """
        aggregator = self.aggregators[metric["kind"]]

        # Compute the new value
        new_data = aggregator.update(key, value, metric["observation"])
        metric["observation"] = new_data

    def finalize_period(self, metric, name, at):
        """ Finalize a metric period. For each registered metric, call the
        finalize method on correspondent aggregator and instantiate a new blank
        metric period.

        Called if either the period time was crossed or can forced on logout.
        """

        # If no data has been gathered
        if metric["observation"] is not None:
            # Retrieve the current period
            # Call the finalize method on the aggregator
            aggregator = self.aggregators[metric["kind"]]

            finished = {}
            finished["observation"] = aggregator.finalize(metric["observation"])
            finished["finish"] = at
            finished["name"] = name
            finished["start"] = metric["start"]
            self.store.append(finished)

        # Reset the period
        metric = self._new_metric(metric["kind"], metric['period'])
        metric["start"] = at
        self.metrics[name] = metric
        return metric

    def get_data_to_publish(self, at):
        """ Return the list of finished periods, reset the list of
        finished periods after.
        """
        for metric_name, metric in self.metrics.items():
            self.finalize_period(metric, metric_name, at)
        finished_periods = self.store
        self.store = []
        return finished_periods

    ###
    # Helpers for debug and tests
    ###

    def list_metrics(self):
        """ Return the list of registered metrics
        """
        return self.metrics.keys()

    def get_metric_kind(self, name):
        """ Return the kind of a given metric
        """
        return self.metrics[name]["kind"]

    def get_metric_period(self, name):
        """ Return the period of a given metric
        """
        return self.metrics[name]["period"]

    def get_metric_start(self, name):
        """ Return the start time for a given metric
        """
        return self.metrics[name]["start"]

    def get_metric_aggregate(self, name):
        """ Return the current aggregated data for a given metric
        """
        return self.metrics[name]["observation"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" High-level interaction with sqreen API
"""
import logging

from .http_client import InvalidStatusCodeResponse, StatusFailedResponse

LOGGER = logging.getLogger(__name__)


class InvalidToken(Exception):
    """ Exception raise when a login fails because of the token value
    """
    pass


class Session(object):
    """ Class responsible for collection date and interacting with the sqreen API
    """

    def __init__(self, connection, api_key):
        self.connection = connection
        self.api_key = api_key
        self.session_token = None

    def login(self, runtime_infos):
        """ Login to the backend
        """
        headers = {'x-api-key': self.api_key}

        try:
            result = self.connection.post('app-login', runtime_infos,
                                          headers=headers,
                                          retries=self.connection.RETRY_LONG)
        except InvalidStatusCodeResponse as exc:
            LOGGER.error("Cannot login. Token may be invalid: %s", self.api_key)
            LOGGER.error("Invalid response: %s", exc.response_data)
            if exc.status in (401, 403):
                raise InvalidToken()
            raise
        except StatusFailedResponse as exc:
            LOGGER.error("Cannot login. Token may be invalid: %s", self.api_key)
            LOGGER.error("Invalid response: %s", exc.response)
            raise InvalidToken()

        LOGGER.debug("Received session_id %s", result['session_id'])
        self.session_token = result['session_id']

        return result

    def is_connected(self):
        """ Return a boolean indicating if a successfull login was made
        """
        return self.session_token is not None

    def _headers(self):
        """ Return authentication headers
        """
        return {'x-session-key': self.session_token}

    def _get(self, url, retries=None):
        """ Call connection.get with right headers
        """
        return self.connection.get(url, headers=self._headers(),
                                   retries=retries)

    def _post(self, url, data, retries=None):
        """ Call connection.get with right headers
        """
        return self.connection.post(url, data, headers=self._headers(),
                                    retries=retries)

    def logout(self):
        """ Logout current instance in the backend
        """
        return self._get('app-logout')

    def heartbeat(self):
        """ Tell the backend that the instance is still up and retrieve latest
        commands
        """
        return self._get('app-beat', retries=self.connection.RETRY)

    def post_attack(self, attack):
        """ Report an attack on the backend
        """
        LOGGER.debug("Post attack %s", attack)
        return self._post('attack', attack, retries=self.connection.RETRY_LONG)

    def post_commands_result(self, commands_result):
        """ Report commands result
        """
        return self._post('commands', commands_result, retries=self.connection.RETRY_LONG)

    def post_sqreen_exception(self, exception):
        """ Report sqreen exception
        """
        return self._post('sqreen_exception', exception, retries=self.connection.RETRY)

    def post_metrics(self, metrics):
        """ Post metrics aggregates to the backend
        """
        # Don't send empty metrics payload
        if len(metrics) < 1:
            return

        data = {'metrics': metrics}
        return self._post('metrics', data, retries=self.connection.RETRY_LONG)

    def get_rulespack(self):
        """ Retrieve rulespack from backend
        """
        return self._get('rulespack', retries=self.connection.RETRY_LONG)

    def post_batch(self, batch):
        """ Post a batch to the backend
        """
        return self._post('batch', {'batch': batch},
                          retries=self.connection.RETRY_LONG)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Various constants
"""

# Constant for callback not changing the workflow
NOTHING = 'NOTHING'
RETRY = 'RETRY'


LIFECYCLE_METHODS = {
    "PRE": "pre",
    "POST": "post",
    "FAILING": "failing"
}

ACTIONS = {
    "RAISE": "raise",
    "OVERRIDE": "override",
    "RETRY": "retry",
    "MODIFY_ARGS": "modify_args"
}


VALID_ACTIONS_PER_LIFECYCLE = {
    LIFECYCLE_METHODS["PRE"]: [
        ACTIONS["RAISE"],
        ACTIONS["OVERRIDE"],
        ACTIONS["MODIFY_ARGS"]],
    LIFECYCLE_METHODS["FAILING"]: [
        ACTIONS["RAISE"],
        ACTIONS["RETRY"],
        ACTIONS["OVERRIDE"]],
    LIFECYCLE_METHODS["POST"]: [
        ACTIONS["RAISE"],
        ACTIONS["OVERRIDE"]
    ]
}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Sqreen Python agent thread composition
"""
import os
import sys
import time
import atexit
import threading
import traceback

from random import randint
from threading import Thread, Event
from logging import getLogger

from sqreen.session import Session, InvalidToken
from sqreen.runtime_infos import RuntimeInfos
from sqreen.http_client import Urllib3Connection
from sqreen.runner import Runner, CappedQueue, RunnerStop, MAX_QUEUE_LENGTH, MAX_OBS_QUEUE_LENGTH
from sqreen.config import Config
from sqreen.remote_command import RemoteCommand
from sqreen.deliverer import SimpleDeliverer
from sqreen.remote_exception import RemoteException
from sqreen.instrumentation import Instrumentation
from sqreen.metrics import MetricsStore
import sqreen.rules_callbacks


RUNNER_THREAD = None

# Graceful timeout, max time waiting for the runner thread to exit before exiting
GRACEFUL_TIMEOUT = 4
GRACEFUL_FAIL_MSG = "Sqreen thread didn't exit after %s seconds"

LOGGER = getLogger(__name__)


# Exit mechanism
def sqreen_exit(runner, queue):
    queue.put(RunnerStop)

    try:
        runner.join(GRACEFUL_TIMEOUT)
    except RuntimeError:
        logger = getLogger(__name__)
        logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)
    else:
        if runner.isAlive():
            logger = getLogger(__name__)
            logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)


def before_hook_point():
    global RUNNER_THREAD
    LOGGER.debug("Runner thread alive %s", RUNNER_THREAD.is_alive())
    if RUNNER_THREAD.is_alive() is False:
        runner = RunnerThread()
        runner.queue = RUNNER_THREAD.queue
        runner.observations_queue = RUNNER_THREAD.observations_queue
        runner.instrumentation = RUNNER_THREAD.instrumentation
        runner.start()

        atexit.register(sqreen_exit, runner=runner, queue=runner.queue)
        RUNNER_THREAD = runner


def _dump_thread_stacks():
    """ Try to dump all the currently running thread stacks
    """
    try:
        msg = ["Threads stacks \n"]
        thread_name_map = {thread.ident: thread.name for thread in threading.enumerate()}

        for thread_id, stack in sys._current_frames().items():
            msg.append("# Thread: id '%r', name %r" % (thread_id, thread_name_map[thread_id]))
            msg.append("".join(traceback.format_stack(stack)))

        LOGGER.debug("\n".join(msg))
    except Exception:
        LOGGER.exception("Something happened")


def start():
    """ Start the background thread and start protection
    """
    # Check if the agent is not disabled first
    try:
        disabled = int(os.getenv('SQREEN_DISABLE', 0))
    except ValueError:
        disabled = None

    # Retrieve the command used to launch the process
    if hasattr(sys, 'argv') and len(sys.argv) >= 1:
        command = sys.argv[0]
    else:
        command = ''

    # Check if we shouldn't launch ourselves
    if 'ipython' in command:
        LOGGER.critical('Sqreen agent is disabled when running ipython.')
        return

    if hasattr(sys, 'argv') and len(sys.argv) >= 2 and sys.argv[1] == 'test':
        LOGGER.critical('Sqreen agent is disabled when running tests.')
        return

    if disabled:
        LOGGER.critical('Sqreen agent is disabled.')
        return

    global RUNNER_THREAD

    # Check for double instrumentation
    if RUNNER_THREAD is None:
        try:
            runner = RunnerThread()
            runner.start()

            atexit.register(sqreen_exit, runner=runner, queue=runner.queue)

            timeout = runner.instrumentation_done.wait(10)

            if timeout is not True:
                msg = "Sqreen couldn't start. Check network connectivity: http://status.sqreen.io/"
                LOGGER.critical(msg)

                # Also dump the stacktraces of all running threads
                _dump_thread_stacks()
            else:
                if runner.isAlive():
                    LOGGER.info("Sqreen instrumentation started successfully")
        except Exception:
            LOGGER.critical("Sqreen thread fails to start, you're not protected",
                            exc_info=True)
            return

        RUNNER_THREAD = runner
    else:
        runner = RUNNER_THREAD

    return runner


class RunnerThread(Thread):
    """ Class responsible for starting the runner and monitor it
    """

    name = "SqreenRunnerThread"

    def __init__(self):
        super(RunnerThread, self).__init__()
        self.daemon = True
        self.logger = getLogger(__name__)
        self.queue = CappedQueue(MAX_QUEUE_LENGTH)
        self.observations_queue = CappedQueue(MAX_OBS_QUEUE_LENGTH)
        self.instrumentation_done = Event()
        self.runtime_infos = RuntimeInfos().all()
        self.instrumentation = Instrumentation(self.queue.put, before_hook_point)

    def run(self):
        """ Launch the runner
        """
        self.logger.debug('Starting Sqreen %s', sqreen.__version__)
        # FIXME this part should be more tested
        while True:
            session = None
            runner = None

            try:
                # Instantiate configuration
                config = Config()
                config.load()

                try:
                    token = config['TOKEN']
                except KeyError:
                    msg = ("Sorry but we couldn't find your Sqreen token.\n"
                           "Your application is NOT currently protected by Sqreen.\n"
                           "\n"
                           "Have you filled your sqreen.ini?")
                    self.logger.critical(msg)
                    self.instrumentation_done.set()
                    return

                self.logger.warning("Using token %s", token)

                # Connection
                url = config['URL']
                self.logger.warning("Connection to %s", url)

                con = Urllib3Connection(url)

                session = Session(con, token)

                try:
                    session.login(self.runtime_infos)
                except InvalidToken:
                    msg = ("Sorry but your Sqreen token appears to be invalid.\n"
                           "Your application is NOT currently protected by Sqreen.\n"
                           "\n"
                           "Please check the token against the interface")
                    self.logger.critical(msg)
                    self.instrumentation_done.set()
                    return

                self.logger.info("Login success")

                deliverer = SimpleDeliverer(session)
                remote_command = RemoteCommand.with_production_commands()
                metrics_store = MetricsStore()
                metrics_store.register_production_aggregators()

                runner = Runner(self.queue, self.observations_queue, session,
                                deliverer, remote_command, self.instrumentation,
                                metrics_store)
                runner.do_heartbeat()
                self.instrumentation_done.set()
                runner.run()
                # If the runner exit, returns
                return
            except Exception:
                self.logger.exception("An unknown exception occured")

                if session is not None and session.is_connected():
                    try:
                        session.post_sqreen_exception(RemoteException(sys.exc_info()).to_dict())
                        session.logout()
                    except Exception:
                        self.logger.exception("Exception while logout")
                        return

                try:
                    self.instrumentation.deinstrument_all()
                except Exception:
                    # We did not managed to remove instrumentation, state is unclear:
                    # terminate thread
                    self.logger.exception("Exception while trying to clean-up")
                    return

                delay = randint(0, 10)
                self.logger.debug("Sleeping %s seconds before retry", delay)
                time.sleep(delay)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Base class for matcher based rules
"""
from logging import getLogger
from collections import Mapping

from ..rules import RuleCallback
from ..matcher import Matcher
from ..exceptions import InvalidArgument

LOGGER = getLogger(__name__)


class MatcherRule(RuleCallback):
    """ Base class for matcher based rules.
    The valid matcher syntax is described here:
    https://github.com/sqreen/Wiki/wiki/String-matcher
    """

    def __init__(self, *args, **kwargs):
        super(MatcherRule, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            raw_patterns = self.data['values']
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        self.matcher = Matcher(raw_patterns)

    def match(self, string):
        """ Check if string match one of rule pattern
        """
        return self.matcher.match(string)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Import hook strategy to prepare hooks for modules
"""
import sys
import logging

from ..helpers import partial
from ..hook_point import hook_point
from ..import_hook import ImportHook, get_hook_parent, get_hook_path

from .base import BaseStrategy

LOGGER = logging.getLogger(__name__)


class ImportHookStrategy(BaseStrategy):
    """ Simple strategy that calls setattr(hook_module, hook_name, callback)
    """

    def __init__(self, hook_name, channel, before_hook_point=None):
        super(ImportHookStrategy, self).__init__(channel, before_hook_point)
        self.hook_module = hook_name

        self.hook_points = set()

    def hook(self, callback):
        """ Accept a callback and store it. If it's the first callback
        for this strategy, actually hook to the endpoint.
        """
        self.add_callback(callback)

        try:
            callback_module_path = callback.hook_module.split('::', 1)[1]
        except IndexError:
            callback_module_path = ''

        self.hook_points.add((callback_module_path, callback.hook_name))

        # Check that we didn't already hooked the endpoint
        if self.hooked is False:
            import_hook = ImportHook(self.hook_module, self.import_hook_callback)
            sys.meta_path.insert(0, import_hook)

            self.hooked = True

    def import_hook_callback(self, module):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module. Called by ImportHook
        """

        for hook_class, hook_name in self.hook_points:
            hook_path = get_hook_path(self.hook_module, hook_class)

            # Retrieve the original
            hook_parent = get_hook_parent(module, hook_class)

            if hook_parent is None:
                msg = "'%s' doesn't exists for module %s at path '%s', couldn't hook"
                LOGGER.warning(msg, hook_class, module, hook_path)
                continue

            original = getattr(hook_parent, hook_name, None)

            # Do not create functions if original don't exists
            if original is None:
                msg = "'%s' doesn't exists for %s at path '%s'"
                LOGGER.warning(msg, hook_name, hook_parent, hook_path)
                continue

            # Hook it
            _hook_point = hook_point(self, hook_path,
                                     hook_name, original)

            # And replace it
            setattr(hook_parent, hook_name, _hook_point)
            LOGGER.debug("Successfully hooking on %s %s", hook_path,
                         hook_name)

        return module

    @staticmethod
    def get_strategy_id(callback):
        """ Return the tuple (callback.hook_module, callback.hook_name) as
        identifier for this strategy
        """
        return callback.hook_module.split('::', 1)[0]

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Look for badly behaved clients
"""
from logging import getLogger

from ..runtime_infos import runtime
from .regexp_rule import RegexpRule


LOGGER = getLogger(__name__)


class UserAgentMatchesCBFramework(RegexpRule):

    def pre(self, original, *args, **kwargs):

        request = runtime.get_current_request()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        user_agent = request.client_user_agent
        if not user_agent:
            return

        match = self.match_regexp(user_agent)
        if not match:
            return

        infos = {'found': match}
        self.record_attack(infos)

        return {'status': 'raise', 'data': match,
                'rule_name': self.rule_name}

UserAgentMatchesCBDjango = UserAgentMatchesCBFramework



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" DBApi2 strategy and custom connection and cursor classes
"""
import sys
import logging
import inspect

from ..helpers import partial, Proxy
from ..hook_point import hook_point
from ..import_hook import ImportHook
from ...utils import update_wrapper

from .base import BaseStrategy


LOGGER = logging.getLogger(__name__)


class CustomCursor(Proxy):
    """ CustomCursor proxy, it proxy a real Cursor of a DBApi2
    module, instrument all methods with prefix ::Cursor.
    """

    def __init__(self, real_cursor, hook_point, module_path):
        super(CustomCursor, self).__init__(real_cursor)

        methods = inspect.getmembers(self._obj, inspect.ismethod)
        builtins = inspect.getmembers(self._obj, inspect.isbuiltin)

        for method in methods + builtins:
            if method[0].startswith('_'):
                # Ignore private or semi-private methods
                continue

            hook_module = "{}::Cursor".format(module_path)
            new_hook_point = hook_point(hook_module, method[0],
                                        method[1])
            object.__setattr__(self, method[0], new_hook_point)


class CustomConnection(Proxy):
    """ CustomConnection proxy, it proxy a real Connection of a DBApi2
    module, instrument all methods with prefix ::Connection and returns
    CustomCursor when callin the cursor method.
    """

    def __init__(self, real_connection, hook_point, module_path):
        self._sqreen_hook_point = hook_point
        self._sqreen_module_path = module_path
        super(CustomConnection, self).__init__(real_connection)

        methods = inspect.getmembers(self._obj, inspect.ismethod)
        builtins = inspect.getmembers(self._obj, inspect.isbuiltin)

        for method in methods + builtins:
            if method[0] == 'cursor':
                # Don't hook on cursor
                continue

            if method[0].startswith('_'):
                # Ignore private or semi-private methods
                continue

            hook_module = "{}::Connection".format(module_path)
            new_hook_point = hook_point(hook_module, method[0], method[1])
            object.__setattr__(self, method[0], new_hook_point)

    def cursor(self, *args, **kwargs):
        """ Instantiate a real cursor, proxy it via CustomCursor.
        """
        return CustomCursor(self._obj.cursor(*args, **kwargs),
                            self._sqreen_hook_point, self._sqreen_module_path)


def custom_connect(hook_point, module_path, original_connect, *args, **kwargs):
    """ Replacement to the connect function of a DBApi2 module. It will
    instantiate a connection via the original connect function and proxy it
    via CustomConnection defined earlier.
    """

    def wrapper(*args, **kwargs):
        return CustomConnection(original_connect(*args, **kwargs), hook_point, module_path)

    wrapper = update_wrapper(wrapper, original_connect)
    return wrapper


class DBApi2Strategy(BaseStrategy):
    """ Strategy for DBApi2 drivers.

    It's different from the SetAttrStrategy and requires special care in
    hook_module and hook_name.
    DBApi2 tries to hook on 'connect' method of DBApi2 compatible driver.
    In order to do so, it needs the module name where 'connect' is available. It
    must be the first part of hook_module, for example in sqlite3, it will be
    'sqlite3'.

    The hook_module could then contains either '::Cursor' for hooking on
    cursor methods or '::Connection' for hooking on connection methods.
    The hook_name will then specify which method it will hook on.

    For example with sqlite3, the tuple ('sqlite3::Connection', 'execute') will
    reference the execute method on a sqlite3 connection.
    In the same way, the tuple ('sqlite3::Cursor', 'execute') will reference
    the execute method on a sqlite3 cursor.

    It will works the same way for all DBApi2 drivers, even with psycopg2
    where cursor class is actually defined as 'psycopg2.extensions.cursor',
    'psycopg2::Cursor' will correctly reference all psycopg2 cursor methods.
    """

    def __init__(self, module_path, channel, before_hook_point=None):
        super(DBApi2Strategy, self).__init__(channel, before_hook_point)
        self.module_path = module_path

        self.hooked = False

    def hook(self, callback):
        """ Accept a callback and store it. If it's the first callback
        for this strategy, actually hook to the endpoint.

        Once hooked, it will instrument all method on connection and cursor.
        But if no callback is defined on a specific method, the overhead will
        be minimal.
        """
        # Register callback
        self.add_callback(callback)

        # Check if we already hooked at
        if not self.hooked:

            import_hook = ImportHook(self.module_path, self.import_hook_callback)
            sys.meta_path.insert(0, import_hook)

            self.hooked = True

    def import_hook_callback(self, module):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module
        """
        original = getattr(module, "connect", None)

        if original is None:
            msg = "'%s' doesn't exists for module %s at path '%s', couldn't hook"
            LOGGER.warning(msg, 'connect', module, self.module_path)
            return

        _hook_point = partial(hook_point, self)
        new_connect = custom_connect(_hook_point, self.module_path, original)
        setattr(module, "connect", new_connect)
        LOGGER.debug("Successfully hooking on %s %s", self.module_path,
                     "connect")

    @staticmethod
    def get_strategy_id(callback):
        """ Returns the module part of hook_module (without everything after
        ::) as identifier for this strategy. Multiple hooks on sqlite3* should
        be done in this strategy.
        """
        # Check if the klass is part module and part klass name
        if '::' in callback.hook_module:
            return callback.hook_module.split('::', 1)[0]
        else:
            return callback.hook_module

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Aggregate authentication tentatives
"""
import logging
import json

from .matcher_callback import MatcherRule
from ..runtime_infos import runtime

LOGGER = logging.getLogger(__name__)


class CustomEncoder(json.JSONEncoder):

    def default(self, obj):
        """ Return the repr of unkown objects
        """
        return repr(obj)


class AuthMetricsCB(MatcherRule):

    def post(self, original, result, *args, **kwargs):
        # Django authentication model return either an User or None
        if result is None:
            auth_status = 'login-fail'
        else:
            auth_status = 'login-success'

        keys = []

        # Search for credentials identifier that match the whitelist
        for identifier in kwargs.keys():
            if self.match(identifier):
                keys.append([identifier.lower(), kwargs[identifier]])

        if not keys:
            # If we couldn't identify an user, don't record a metric
            return

        request = runtime.get_current_request()
        key = {'keys': keys, 'ip': request.client_ip}

        observation_key = json.dumps(key, separators=(',', ':'), sort_keys=True, cls=CustomEncoder)
        self.record_observation(auth_status, observation_key, 1)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Base class for JS Callbacks
"""
import logging

from py_mini_racer import py_mini_racer
from py_mini_racer.py_mini_racer import JSEvalException

from ..rules import RuleCallback
from ..binding_accessor import BindingAccessor
from ..runtime_infos import runtime
from ..exceptions import SqreenException
from ..frameworks.blank import BlankRequest


LOGGER = logging.getLogger(__name__)


class JSException(SqreenException):
    """ Base exception raised in JSCB
    """
    def __init__(self, message, callback, arguments):
        super(JSException, self).__init__(message)
        self.callback = callback
        self.arguments = arguments

    def exception_infos(self):
        return {'cb': self.callback, 'args': self.arguments}


class JSCB(RuleCallback):
    """ A callback that run a JS function as pre / post / failing through
    py_mini_racer context.
    """

    def __init__(self, *args, **kwargs):
        super(JSCB, self).__init__(*args, **kwargs)

        # Prepare a js context
        self.js_context = py_mini_racer.MiniRacer()

        self.arguments = {}

        if self.callbacks:
            for callback_name, callback_args in self.callbacks.items():

                if not isinstance(callback_args, list):
                    source = callback_args
                    self.arguments[callback_name] = []
                else:
                    source = callback_args[-1]
                    arguments = callback_args[:-1]
                    self.arguments[callback_name] = [BindingAccessor(arg) for arg in arguments]

                js_source = "var {} = {}".format(callback_name, source)
                self.js_context.eval(js_source)

    def __getattribute__(self, name):
        """ Lie about pre / post / failing existence if no callbacks is defined
        for them
        """
        if name in ('pre', 'post', 'failing'):
            if name not in self.callbacks:
                err_msg = "'{}' object has no attribute '{}'"
                raise AttributeError(err_msg.format(self.__class__.__name__, name))

        return RuleCallback.__getattribute__(self, name)

    def pre(self, original, *args, **kwargs):
        """ Call the pre JS function with its arguments
        """
        return self.execute('pre', self.arguments['pre'], original, None, args, kwargs)

    def post(self, original, return_value, *args, **kwargs):
        """ Call the post JS function with its arguments
        """
        return self.execute('post', self.arguments['post'], original, return_value, args, kwargs)

    def failing(self, original, exception, *args, **kwargs):
        """ Call the failing JS function with its arguments
        """
        return self.execute('failing', self.arguments['failing'], original, exception, args, kwargs)

    def execute(self, name, arguments, original, return_value, args, kwargs):
        """ Execute a JS callback passed in definition.
        Handle recording attack, observations and chaining.
        Protected against infinite recursion with a max number of JS calls
        set to 100.
        """

        request = runtime.get_current_request()

        # Fallback on a blank request for binding accessor
        if request is None:
            request = BlankRequest()

        # Safeguard against infinite recursion
        for i in range(100):
            binding_eval_args = {
                "binding": locals(),
                "global_binding": globals(),
                "framework": request,
                "instance": original,
                "arguments": args,
                "kwarguments": kwargs,
                "cbdata": self.data,
                "return_value": return_value
            }
            resolved_args = [arg.resolve(**binding_eval_args) for arg in arguments]

            try:
                result = self.js_context.call(name, *resolved_args)
            except JSEvalException as e:
                raise JSException(e.args[0], name, resolved_args)

            LOGGER.debug("JS Result %r for %s", result, self.rule_name)

            if result is None:
                return result

            # Process the return value
            self._record_attack(result)
            self._record_observations(result)

            # Check for chaining
            if result.get('call') is None:
                return result

            # Prepare next call
            name = result['call']

            if name not in self.callbacks:
                raise JSException("Invalid callback '{}'".format(name), name, None)

            return_value = result.get('data')

            if result.get('args'):
                arguments = [BindingAccessor(arg) for arg in result['args']]
            else:
                arguments = self.arguments[name]

    def _record_attack(self, return_value):
        """ Record an attack if the JS callback returned a record info
        """
        if return_value.get('record'):
            self.record_attack(return_value['record'])

    def _record_observations(self, return_value):
        """ Record observations if the JS callback returned a observations list
        """
        if return_value.get('observations'):
            for observation in return_value['observations']:
                self.record_observation(*observation)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Look for badly behaved clients
"""
from logging import getLogger

from ..rules import RuleCallback
from ..runtime_infos import runtime


LOGGER = getLogger(__name__)


class NotFoundCBFlask(RuleCallback):

    def post(self, original, response, *args, **kwargs):
        if response.status_code == 404:
            current_request = runtime.get_current_request()

            # A 404 may prevent record request context to store the request
            # store it without arguments
            if current_request is None:
                LOGGER.warning("No request was recorded abort")
                return

            infos = {
                'path': current_request.path,
                'host': current_request.hostname,
                'verb': current_request.method,
                'ua': current_request.client_user_agent
            }
            self.record_attack(infos)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Module called when sitecustomize is automatically imported by python
"""
import sys
from sqreen import start

# Try to detect when this module is imported as sitecustomize
if not hasattr(sys, 'argv'):
    start()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Hijacking strategies
"""
import logging

from importlib import import_module

from ..hook_point import hook_point
from ...exceptions import SqreenException

from .base import BaseStrategy
from .dbapi2 import DBApi2Strategy
from .import_hook import ImportHookStrategy
from .psycopg2 import Psycopg2Strategy
from .flask import FlaskStrategy
from .django import DjangoStrategy

__all__ = ['BaseStrategy', 'DBApi2Strategy', 'ImportHookStrategy',
           'Psycopg2Strategy', 'FlaskStrategy', 'DjangoStrategy']

LOGGER = logging.getLogger(__name__)


class InvalidHookPoint(SqreenException):
    """ Exception raised when the hook_point couldn't be found or is invalid
    """
    pass


class SetAttrStrategy(BaseStrategy):
    """ Simple strategy that calls setattr(hook_module, hook_name, callback)
    """

    def __init__(self, hook_name, channel, before_hook_point=None):
        super(SetAttrStrategy, self).__init__(channel, before_hook_point)
        self.hook_module, self.hook_name = hook_name
        self.original = self._get_original()

    def hook(self, callback):
        """ Accept a callback and store it. If it's the first callback
        for this strategy, actually hook to the endpoint.
        """
        self.add_callback(callback)

        # Check that we didn't already hooked the endpoint
        if self.hooked is False:
            _hook_point = hook_point(self, callback.hook_module,
                                     callback.hook_name, self.original)

            module = self._get_hook_to()
            setattr(module, self.hook_name, _hook_point)

            self.hooked = True

    def _get_original(self):
        """ Return the original function defined at (hook_module, hook_name)
        """

        if self.hook_name == '':
            raise InvalidHookPoint('Empty hook_name')

        module = self._get_hook_to()

        try:
            return getattr(module, self.hook_name)
        except AttributeError:
            msg = "Bad hook_name {} on hook_module {}"
            raise InvalidHookPoint(msg.format(self.hook_name,
                                              self.hook_module))

    def _get_hook_to(self):
        """ Retrieve the python module or class corresponding to hook_name
        """

        if self.hook_module == '':
            raise InvalidHookPoint('Empty hook_module')

        # Check if the klass is part module and part klass name
        if '::' in self.hook_module:
            module_name, class_name = self.hook_module.split('::', 1)
        else:
            module_name, class_name = self.hook_module, None

        try:
            module = import_module(module_name)
        except ImportError:
            raise InvalidHookPoint("Bad module name {}".format(module_name))

        if class_name:
            try:
                module = getattr(module, class_name)
            except AttributeError:
                msg = "Bad class_name {} on module {}".format(class_name, module_name)
                raise InvalidHookPoint(msg)

        return module

    @staticmethod
    def get_strategy_id(callback):
        """ Return the tuple (callback.hook_module, callback.hook_name) as
        identifier for this strategy
        """
        return (callback.hook_module, callback.hook_name)

    def _restore(self):
        """ Restore the original function at the hooked path
        """
        module = self._get_hook_to()
        setattr(module, self.hook_name, self.original)
        self.hooked = False



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Base strategy
"""
import logging

from collections import defaultdict

from ..hook_point import LIFECYCLE_METHODS

LOGGER = logging.getLogger(__name__)


class BaseStrategy(object):
    """ The base strategy holds callbacks for a hook point

    Subclasses need to define way to hook and deinstrument.
    BaseStrategy accepts a channel, usually passed from Instrumentation
    directly.
    """

    def __init__(self, channel, before_hook_point=None):
        self.unique_callbacks = set()
        # Data structure for {[hook_module, hook_name]: {pre: [], post: [], fail: []}}
        self.callbacks = defaultdict(lambda: defaultdict(list))

        self.hooked = False

        self.channel = channel
        self._before_hook_point = before_hook_point

    def add_callback(self, callback):
        """ Add a callback on a specific endpoint defined by the couple
        (hook_module, hook_name) embedded in callback
        """
        key = (callback.hook_module, callback.hook_name)

        for method in LIFECYCLE_METHODS.values():
            if hasattr(callback, method):
                self.callbacks[key][method].append(callback)

        self.unique_callbacks.add(callback)

    def get_pre_callbacks(self, key):
        """ Returns callbacks with a pre lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS['PRE']]

    def get_post_callbacks(self, key):
        """ Returns callbacks with a post lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS['POST']]

    def get_failing_callbacks(self, key):
        """ Returns callbacks with a failing lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS['FAILING']]

    def _total_callbacks(self):
        """ Count the total number of callbacks
        """
        return len(self.unique_callbacks)

    def deinstrument(self, callback):
        """ Deactivate the callback for this endpoint.
        If it was the latests callback for this endpoint, restore the original
        hooked function.
        """
        key = (callback.hook_module, callback.hook_name)

        callbacks = self.callbacks[key]

        LOGGER.debug("Deinstrumenting %s", callback)

        for method in LIFECYCLE_METHODS.values():
            if hasattr(callback, method):
                try:
                    callbacks[method].remove(callback)
                except ValueError:
                    # Callback has not been added in this strategy
                    msg = "Callback %s %s trying to be deinstrumented but not instrumented"
                    LOGGER.warning(msg, method, callback)

        try:
            self.unique_callbacks.remove(callback)
        except KeyError:
            pass

        LOGGER.debug("Number of remaining callback for %s: %s", self, self._total_callbacks())

        # If no more callbacks are set deinstrument
        if self._total_callbacks() is 0:
            self._restore()

    def deinstrument_all(self):
        """ Deinstrument all callbacks
        """
        for callback in list(self.unique_callbacks):
            self.deinstrument(callback)

        self.unique_callbacks = set()

    def _restore(self):
        """ Restore the original method, subclasses need to overload this method
        """
        self.hooked = False

    def before_hook_point(self):
        """ Run code just before running a hook_point
        """
        if self._before_hook_point is not None:
            self._before_hook_point()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Custom strategy for psycopg2 oddities
"""
import logging

from .dbapi2 import CustomConnection, CustomCursor
from .import_hook import ImportHookStrategy


LOGGER = logging.getLogger(__name__)


def wrap_custom_register_type(original):

    def custom_register_type(obj, scope=None):
        LOGGER.debug("register_type %s", locals())
        if isinstance(scope, (CustomConnection, CustomCursor)):
            scope = scope._obj

        return original(obj, scope)

    return custom_register_type


class Psycopg2Strategy(ImportHookStrategy):
    """ Simple strategy that replace psycopg2.*.register_type to works with our
    Proxy objects.
    """

    def import_hook_callback(self, module):
        """ Monkey-patch the object located at psycopg2.*.register_type to
        monkey-patch the register_type function
        """

        if hasattr(module, 'register_type'):
            wrapped = wrap_custom_register_type(module.register_type)
            setattr(module, 'register_type', wrapped)

            LOGGER.debug("Successfully hooking on %s %s", self.hook_module,
                         "register_type")
        else:
            LOGGER.debug("Couldn't hook on %s", self.hook_module)

        return module



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Sqreen exceptions
"""


class SqreenException(Exception):
    """ Base exception for all sqreen exceptions
    """

    def __str__(self):
        return self.__repr__()

    def exception_infos(self):
        return {}


class InvalidArgument(SqreenException):
    """ Exception raised when sqreen code receive invalid arguments like bad
    rule dict.
    """
    pass


# This exception name is particularly important since it is often seen by
# Sqreen users when watching their logs. It should not raise any concern to
# them.
class AttackBlocked(SqreenException):
    """ Raised when a callback detected an attack
    """

    def __init__(self, rule_name):
        self.rule_name = rule_name

    def __repr__(self):
        msg = "Sqreen blocked a security threat (type: #{}). No action is required."
        return msg.format(self.rule_name)

###
# HTTP Exceptions
###


class SqreenHttpException(SqreenException):
    pass


class InvalidResponseContentType(SqreenHttpException):

    def __init__(self, content_type):
        self.content_type = content_type

    def __repr__(self):
        return 'Invalid response Content-Type: {!r}'.format(self.content_type)


class InvalidJsonResponse(SqreenHttpException):

    def __init__(self, parsing_exception):
        self.parsing_exception = parsing_exception

    def __repr__(self):
        msg = 'An error occured while trying to parse the response: {!r}'
        return msg.format(self.parsing_exception)


class StatusFailedResponse(SqreenHttpException):

    def __init__(self, response):
        self.response = response

    def __repr__(self):
        msg = 'Response returned with a status false: {!r}'
        return msg.format(self.response)


class InvalidStatusCodeResponse(SqreenHttpException):

    def __init__(self, status, response_data=None):
        self.status = status
        self.response_data = response_data

    def __repr__(self):
        msg = 'Response status code is invalid: {!r}'
        return msg.format(self.status)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Main runner module
"""

import logging
from time import time
from datetime import datetime

try:
    # Python2
    from queue import Empty, Queue, Full
except ImportError:
    from Queue import Empty, Queue, Full

MAX_QUEUE_LENGTH = 100
MAX_OBS_QUEUE_LENGTH = 1000


class RunnerStop(object):
    """ Placeholder event for asking the runner to stop
    """


class MetricsEvent(object):
    """ Placeholder for asking observations aggregation to run
    """


class CappedQueue(object):
    """ Capped queue with opiniatied methods
    """

    def __init__(self, maxsize=None):
        if maxsize is None:
            maxsize = MAX_QUEUE_LENGTH

        self.maxsize = maxsize
        self.queue = Queue(self.maxsize)

    def get(self, timeout):
        """ Wait for up to timeout for an item to return and block while
        waiting
        """
        return self.queue.get(block=True, timeout=timeout)

    def get_nowait(self):
        """ Get without waiting, raise queue.Empty if nothing is present
        """
        return self.queue.get_nowait()

    def put(self, item):
        """ Tries to put an item to the queue, if the queue is empty, pop an
        item and try again
        """
        pushed = False
        while pushed is False:
            try:
                self.queue.put_nowait(item)
                pushed = True
            except Full:
                try:
                    self.queue.get_nowait()
                    self.queue.task_done()
                except Empty:
                    pass

    def clear(self):
        """ Clear all items in the queue
        """
        with self.queue.mutex:
            self.queue.queue.clear()

    def half_full(self):
        """ Return True if the current queue size if at least half the maxsize
        """
        return self.queue.qsize() > (self.maxsize / 2.)


class Runner(object):
    """ Main runner class

    Its job is to be the orchestrator and receiver for application communication
    It interacts with the backend through session, call heartbeat himself,
    execute commands and forward events
    """

    # At start, heartbeat is every 15 seconds
    HEARTBEAT_INITIAL_DELAY = 15
    # During one hour
    HEARTBEAT_WARMUP = 60 * 60
    # Then delay raises to 5 minutes
    HEARTBEAT_MAX_DELAY = 5 * 60

    def __init__(self, queue, observation_queue, session, deliverer,
                 remote_command, instrumentation, metrics_store):
        self.logger = logging.getLogger('{}.{}'.format(self.__module__, self.__class__.__name__))
        self.queue = queue
        self.observation_queue = observation_queue
        self.deliverer = deliverer
        self.remote_command = remote_command
        self.instrumentation = instrumentation
        self.metrics_store = metrics_store
        self.stop = False

        # Save the time runner started for checking warmup period termination
        self.started = time()

        # The first time we shouldn't wait too long before sending heartbeat
        self.heartbeat_delay = self.HEARTBEAT_INITIAL_DELAY
        self.publish_metrics_delay = self.HEARTBEAT_MAX_DELAY

        self.last_heartbeat_request = 0
        self.last_post_metrics_request = time()

        self.session = session

    def run(self):
        """ Infinite loop
        """
        self.logger.debug("Starting runner")
        while self.stop is False:
            self.run_once()
        self.logger.debug("Exiting now")

    def run_once(self):
        """ Tries to pop a message or send an heartbeat
        """
        try:
            event = self.queue.get(timeout=self.sleep_delay)
            self.logger.debug('Run once, event: %s after %ss sleep delay', event, self.sleep_delay)
            self.handle_message(event)

            # Exit now if should stop
            if self.stop:
                return
        except Empty:
            self.logger.debug('No message after %ss sleep delay', self.sleep_delay)

        if self._should_do_heartbeat():
            self.do_heartbeat()

        # Aggregate observations in transit in observations queue
        self.aggregate_observations()

        if self._should_publish_metrics():
            self.publish_metrics()

        # Tick the deliverer to publish batch if necessary
        self.deliverer.tick()

    @property
    def sleep_delay(self):
        """ Compute the sleeping delay by taking the heartbeat and metrics delay
        minimum
        """
        return min([self.heartbeat_delay, self.publish_metrics_delay])

    def handle_message(self, event):
        """ Handle incoming message
        Process RunnerStop message or pass event to the deliverer
        """
        if event is RunnerStop:
            self.logger.debug('RunnerStop found, logout')
            self.logout()
            self.stop = True
        elif event is MetricsEvent:
            self.aggregate_observations()
        else:
            self.deliverer.post_event(event)

    def process_commands(self, commands):
        """ handle commands
        """
        while len(commands) > 0:
            result = self.remote_command.process_list(commands, self)
            result = self.session.post_commands_result(result)
            commands = result['commands']

    def do_heartbeat(self):
        """ Do an heartbeat
        """
        res = self.session.heartbeat()
        self.last_heartbeat_request = time()
        self._update_heartbeat_delay()
        self.process_commands(res['commands'])

    def publish_metrics(self):
        """ Publish finished metrics from MetricsStore
        """
        self.last_post_metrics_request = time()
        self.session.post_metrics(self.metrics_store.get_data_to_publish(datetime.utcnow()))

    def aggregate_observations(self):
        """ Empty the observation queue and update the metric store
        with the observations in the queue.
        """
        try:
            while True:
                observation = self.observation_queue.get_nowait()
                self.metrics_store.update(*observation)
        except Empty:
            pass

    def _update_heartbeat_delay(self):
        """ Update the heartbeat_delay if warmup period is finished
        """
        if time() - self.started > self.HEARTBEAT_WARMUP:
            self.heartbeat_delay = self.HEARTBEAT_MAX_DELAY

    def _should_do_heartbeat(self):
        """ Check if we should send an heartbeat because the delay is overdue
        """
        return (self.last_heartbeat_request + self.heartbeat_delay) < time()

    def _should_publish_metrics(self):
        """ Check if we should publish the metrics because the delay is overdue
        """
        return (self.last_post_metrics_request + self.publish_metrics_delay) < time()

    def logout(self):
        """ Run cleanup
        """
        self.logger.debug("Logout")

        # Flush metrics
        self.aggregate_observations()
        self.publish_metrics()

        # Drain deliverer
        self.deliverer.drain()

        self.session.logout()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Matcher module
"""

import re
from operator import eq
from functools import partial

from .utils import is_unicode


MATCHERS = {
    'anywhere': lambda pattern, value: pattern in value,
    'starts_with': lambda pattern, value: value.startswith(pattern),
    'ends_with': lambda pattern, value: value.endswith(pattern),
    'equals': eq
}


class Matcher(object):
    """ Fast regex-like matcher
    """

    def __init__(self, patterns):
        self.insensitives_string_patterns = []
        self.sensitives_string_patterns = []
        self.regex_patterns = []

        self._prepare(patterns)

    def _prepare(self, patterns):
        """ Prepare all the patterns
        """
        for pattern in patterns:

            pattern_type = pattern['type']
            value = pattern['value']
            case_sensitive = pattern.get('case_sensitive', False)

            if pattern_type == 'string':
                pattern_options = pattern.get('options')

                # Get the match function name, defaulting to anywhere
                if not pattern_options:
                    match_name = 'anywhere'
                else:
                    match_name = pattern_options[0]

                if match_name not in MATCHERS.keys():
                    raise ValueError('Unknown match function {}'.format(match_name))

                # Lower the value in case of case insensitive match or
                # for not lower case pattern, it would never match
                if case_sensitive is False:
                    value = value.lower()

                match = partial(MATCHERS[match_name], value)

                if case_sensitive is False:
                    self.insensitives_string_patterns.append(match)
                else:
                    self.sensitives_string_patterns.append(match)

            elif pattern_type == 'regexp':
                if case_sensitive:
                    flags = 0
                else:
                    flags = re.IGNORECASE

                if 'multiline' in pattern.get('options', []):
                    flags = re.MULTILINE | flags

                self.regex_patterns.append(re.compile(value, flags))
            else:
                raise ValueError('Unknown pattern type %s', pattern_type)

    def match(self, value):
        """ Check if string value match one of the string or regex pattern the fastest
        way possible. Accept python 2 str, unicode and python 3 bytes and str.
        """

        # Correct encoding if possible and needed
        if not is_unicode(value):
            value = value.decode('utf-8', errors='replace')

        # Match case sensitive values
        for string_match in self.sensitives_string_patterns:
            if string_match(value):
                return True

        insensitive_value = value.lower()

        # Match case insensitive values
        for string_match in self.insensitives_string_patterns:
            if string_match(insensitive_value):
                return True

        # Match regexes
        for pattern_regex in self.regex_patterns:
            if pattern_regex.search(value) is not None:
                return True

        return False



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Record the current request in flask application
"""
from logging import getLogger

from ..rules import RuleCallback
from ..runtime_infos import runtime
from ..frameworks.flask import FlaskRequest


LOGGER = getLogger(__name__)


class RecordRequestContextFlask(RuleCallback):

    @staticmethod
    def pre(original, *args, **kwargs):
        from flask import request
        runtime.store_request(FlaskRequest(request))

    @staticmethod
    def post(*args, **kwargs):
        runtime.clear_request()

    @staticmethod
    def failing(*args, **kwargs):
        """ Post is always called in a Flask Middleware, don't clean the
        request right now as it may be needed in a post callback
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Environment helpers
"""

import os


def get(name, default=None):
    """ Alias for os.getenv
    """
    return os.getenv(name, default=default)


def get_env_or_fail(name):
    """ Try to get environment variable via os.getenv.
    If it's not defined, raise a ValueError
    """
    res = os.getenv(name)
    if res is None:
        raise ValueError("environment variable %r is not defined" % name)
    else:
        return res



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Logging module helpers
"""
import logging

ROOT_LOGGER_NAME = 'sqreen'
LOG_FORMAT = ('[%(levelname)s][%(asctime)s #%(process)d.%(threadName)s]'
              ' %(name)s:%(lineno)s \t%(message)s')


def configure_root_logger(log_level, log_location=None):
    """ Configure the sqreen root logger. Set following settings:

    - log_level

    Ensure that the sqreen root logger don't propagate messages logs
    to the python root logger.
    Configure two handlers, one stream handler on stderr for errors
    and one file handler if log_location is set for configured level
    """
    logger = logging.getLogger(ROOT_LOGGER_NAME)

    # Don't propagate messages to upper loggers
    logger.propagate = False

    formatter = logging.Formatter(LOG_FORMAT)

    # Only add handler once
    if len(logger.handlers) is 0:

        # Configure the stderr handler configured on CRITICAL level
        stderr_handler = logging.StreamHandler()
        stderr_handler.setFormatter(formatter)
        logger.addHandler(stderr_handler)

        if log_location is not None:
            try:
                handler = logging.FileHandler(log_location)

                handler.setFormatter(formatter)
                logger.addHandler(handler)
            except (OSError, IOError):
                msg = "Couldn't use %s as sqreen log location, fallback to stderr."
                logger.exception(msg, log_location)

    try:
        logger.setLevel(log_level)
    except ValueError:
        logger.error("Unknown log_level %r, don't alter log level", log_level)

    return logger



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Insert X-Protected-By header
"""
import logging
from collections import Mapping

from ..rules import RuleCallback
from ..exceptions import InvalidArgument


LOGGER = logging.getLogger(__name__)


def convert_to_str(headers):
    """ Encode a list of headers tuples into latin1
    """
    for header_name, header_value in headers:
        header_name = str(header_name.encode('latin-1', errors='replace').decode('latin-1'))
        header_value = str(header_value.encode('latin-1', errors='replace').decode('latin-1'))
        yield (header_name, header_value)


class HeadersInsertCBFlask(RuleCallback):
    """ Callback that add the custom sqreen header
    """

    def __init__(self, *args, **kwargs):
        super(HeadersInsertCBFlask, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            self.values = self.data['values']
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        self._headers = None

    @property
    def headers(self):
        """ Cached property to defer headers data conversion
        """
        if self._headers is None:
            self._headers = dict(convert_to_str(self.values))
        return self._headers

    def post(self, original, response, *args, **kwargs):
        """ Set headers
        """
        try:
            for header_name, header_value in self.headers.items():
                response.headers.set(header_name, header_value)
        except Exception:
            LOGGER.warning("An error occured", exc_info=True)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Base class for regexp based rules
"""
import re
from logging import getLogger
from collections import Mapping

from ..rules import RuleCallback
from ..exceptions import InvalidArgument

LOGGER = getLogger(__name__)


class RegexpRule(RuleCallback):
    """ Base class for callback using regexp data to detect attacks.
    Provides the match_regexp method to quickly check if a string match
    one of the regexp data values.
    """

    def __init__(self, *args, **kwargs):
        super(RegexpRule, self).__init__(*args, **kwargs)

        self.patterns = []
        self._prepare_patterns()

    def _prepare_patterns(self):
        """ Precompile regexp patterns
        """
        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            raw_patterns = self.data['values']
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        for pattern in raw_patterns:
            try:
                self.patterns.append(re.compile(pattern, re.IGNORECASE))
            except (re.error, AssertionError):
                LOGGER.warning("%s fails to compile", pattern, exc_info=True)

    def match_regexp(self, string):
        """ Check if string match one of rule pattern
        """
        for pattern in self.patterns:
            if pattern.match(string):
                # Returns the string pattern that matched
                return pattern.pattern



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Sqreen python agent package
"""
from sqreen.config import Config
from sqreen.log import configure_root_logger
from sqreen.runner_thread import start

__author__ = 'Boris Feld'
__email__ = 'boris@sqreen.io'
__version__ = '0.8.7'
VERSION = __version__

# Configure logging
config = Config()
config.load()
configure_root_logger(config['LOG_LEVEL'], config['LOG_LOCATION'])

__all__ = ['start']



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Remote exception module
"""
import traceback

from .event import Event


def traceback_formatter(backtrace):
    """ Accept a backtrace in the format of traceback.extract_tb or
    traceback.extract_stack and returns a list of dictionary matching this
    format::

    {
        'file': FILENAME,
        'line_number': LINE_NUMBER,
        'method': FUNCTION_NAME
    }
    """
    frames = []
    for frame in backtrace:
        filename, line_number, function_name, _ = frame
        frames.append({
            'file': filename,
            'line_number': line_number,
            'method': function_name
        })
    return frames


def raw_traceback_formatter(raw_backtrace):
    """ Accept a traceback object, convert it to a traceback and returns the
    same format than backtrack_formatter.
    """
    return traceback_formatter(traceback.extract_tb(raw_backtrace))


class RemoteException(Event):

    def __init__(self, exc_info, callback_payload=None, exception_payload=None,
                 request_payload=None):
        self.exception_class = exc_info[0].__name__
        self.exception_msg = str(exc_info[1])
        self.raw_backtrace = exc_info[2]
        self.backtrace = raw_traceback_formatter(self.raw_backtrace)

        if callback_payload is None:
            callback_payload = {}
        self.callback_payload = callback_payload

        self.exception_payload = exception_payload
        self.request_payload = request_payload

    def to_dict(self):
        """ Returns information about exception, backtrace and request merged
        into initial payload
        """
        base_payload = {'infos': {}}

        # Base fields
        base_payload['rule_name'] = self.callback_payload.pop('rule_name', None)
        base_payload['rulespack_id'] = self.callback_payload.pop('rulespack_id', None)
        base_payload['rule_signature'] = self.callback_payload.pop('rule_signature', None)

        if self.callback_payload:
            base_payload['infos']['callback'] = self.callback_payload

        if self.exception_payload:
            base_payload['infos']['exception'] = self.exception_payload

        if self.request_payload:
            base_payload.update(self.request_payload)

        base_payload.update({
            'klass': self.exception_class,
            'message': self.exception_msg,
            'context': {'backtrace': self.backtrace}
        })

        return base_payload



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Import hook related helpers
"""
import sys
import imp
import logging

LOGGER = logging.getLogger(__name__)


def get_hook_parent(module, hook_class):
    """ From a module, its name and a class, retrieve the right class
    and returns both the hook parent and the full hook path
    """
    if hook_class:
        if '.' in hook_class:
            raise NotImplementedError()
        else:
            return getattr(module, hook_class, None)
    else:
        return module


def get_hook_path(hook_module, hook_class):
    """ Return the full hook path from hook_module and potentially None hook_class
    """
    if hook_class:
        return '{}::{}'.format(hook_module, hook_class)
    else:
        return hook_module


class ImportHook(object):
    """ Custom import hook that import a module and hook on several
    hook points of this module
    """

    def __init__(self, module_name, hook_callback):
        self.module_name = module_name
        self.hook_callback = hook_callback
        self.executing = False
        self.path = None

        # If modules was already imported, try to hook on them
        if self.module_name in sys.modules:
            msg = "Module %s was already imported, hooking now"
            LOGGER.warning(msg, self.module_name)
            self.hook_callback(sys.modules[self.module_name])

    def find_module(self, fullname, path=None):
        """ Check if loading module name match the one we want
        """
        if fullname == self.module_name:
            self.path = path
            return self
        return

    def load_module(self, name):
        """ Import and hook the module
        """
        LOGGER.debug("Trying to monkey-patch module %s", name)
        # Hierachical import
        module = self._load_module(name, self.path)

        try:
            self.hook_callback(module)
        except Exception:
            err_msg = "Error while hooking on module %s"
            LOGGER.warning(err_msg, self.module_name, exc_info=True)
        finally:
            sys.modules[name] = module
            return module

    @staticmethod
    def _load_module(name, path):
        """ Load a module without putting it in sys.modules
        """
        if name in sys.modules:
            LOGGER.warning("Module %s was already imported", name)
            return sys.modules[name]
        else:
            splitted_name = name.split('.')

            submodule = None

            # The algorithm for loading path inside a package
            # is to load the package first then the subpackage until to the
            # module.
            # find_module take the package name, eg: 'app' for flask.app
            # load_module take the full name, eg 'flask.app' for flask.app
            for index, part in enumerate(splitted_name):
                subpath = '.'.join(splitted_name[:index + 1])

                if subpath in sys.modules:
                    submodule = sys.modules[subpath]
                else:
                    src_file = None
                    try:
                        src_file, pathname, description = imp.find_module(part, path)
                        submodule = imp.load_module(subpath, src_file, pathname,
                                                    description)
                    finally:
                        if src_file:
                            src_file.close()

                path = getattr(submodule, '__path__', None)

            return submodule

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.module_name,
                               self.hook_callback)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Low-level HTTP interaction with sqreen API
"""
import os
import json
from json import JSONEncoder
from datetime import datetime

import sqreen

from .exceptions import InvalidResponseContentType, InvalidJsonResponse
from .exceptions import StatusFailedResponse, InvalidStatusCodeResponse

try:
    # Python 2
    import urlparse
except ImportError:
    import urllib.parse as urlparse

HAS_URLLIB3 = True
try:
    from urllib3 import Retry, PoolManager
    from urllib3.util.request import make_headers
except ImportError:
    try:
        from sqreen.vendors.urllib3 import Retry, PoolManager
        from sqreen.vendors.urllib3.util.request import make_headers
    except ImportError:
        HAS_URLLIB3 = False


class SqreenJsonEncoder(JSONEncoder):
    """ Custom JsonEncoder which can handle datetime objects
    >>> from datetime import datetime
    >>> date = datetime(2016, 3, 22, 14, 43, 14, 742306)
    >>> json.dumps(date, cls=SqreenJsonEncoder)
    '"2016-03-22T14:43:14.742306"'
    >>> json.dumps(2 + 1j, cls=SqreenJsonEncoder)
    '"(2+1j)"'
    >>> class Fail(object):
    ...     def __repr__(self):
    ...         raise NotImplementedError("Oups")
    ...
    >>> f = Fail()
    >>> json.dumps(f, cls=SqreenJsonEncoder)
    '"instance of type <class \\'sqreen.http_client.Fail\\'>"'
    """

    def default(self, obj):  # pylint: disable=E0202

        if isinstance(obj, datetime):
            return obj.isoformat()

        # For unknows types, JSONEncoder will raise a TypeError, instead,
        # returns obj repr
        try:
            return repr(obj)
        # If failing, return a fallback string here with the class name
        except Exception:
            return "instance of type {}".format(repr(obj.__class__))


def where_crt():
    """ Returns the path of the bundled crt
    """
    current_file = os.path.split(__file__)[0]
    return os.path.join(current_file, 'ca.crt')


class Urllib3Connection(object):
    """ Class responsible for making http request to sqreen API,

    handle connection pooling, retry and inbound/outbound formatting
    """
    RETRY_CONNECT_SECONDS = 10
    RETRY_REQUEST_SECONDS = 10

    MAX_DELAY = 60 * 30

    RETRY_STATUS = {503, 504}
    RETRY = Retry(total=5, method_whitelist=False,
                  status_forcelist=RETRY_STATUS, backoff_factor=1)
    RETRY_LONG = Retry(total=128, method_whitelist=False,
                       status_forcelist=RETRY_STATUS, backoff_factor=1)

    PATH_PREFIX = "/sqreen/v0/"

    def __init__(self, server_url):
        self.server_url = server_url
        self.parsed_server_url = urlparse.urlparse(server_url)

        if HAS_URLLIB3 is False:
            raise NotImplementedError("Couldn't find urllib3")
        self.connection = PoolManager(cert_reqs='CERT_REQUIRED',
                                      ca_certs=where_crt())

        user_agent = "sqreen/py-{}".format(sqreen.__version__)
        self.base_headers = make_headers(keep_alive=True, accept_encoding=True,
                                         user_agent=user_agent)

    def _url(self, url):
        """ Format and return an url based on instance path prefix
        """
        path = urlparse.urljoin(self.PATH_PREFIX, url)
        return urlparse.urlunparse((self.parsed_server_url[
            0], self.parsed_server_url[1], path, None, None, None))

    def post(self, endpoint, data=None, headers=None, retries=None):
        """ Post a request to the backend
        """
        url = self._url(endpoint)

        if headers is None:
            headers = {}

        if data is not None:
            data = json.dumps(data, separators=(',', ':'), cls=SqreenJsonEncoder)
            headers['Content-Type'] = 'application/json'

        # Add base headers
        headers.update(self.base_headers)

        response = self.connection.urlopen('POST',
                                           url,
                                           headers=headers,
                                           body=data,
                                           retries=retries)

        return self._parse_response(response)

    def get(self, endpoint, headers=None, retries=None):
        """ Get an endpoint in the backend
        """
        url = self._url(endpoint)

        if headers is None:
            headers = {}

        # Add base headers
        headers.update(self.base_headers)

        response = self.connection.urlopen('GET',
                                           url,
                                           headers=headers,
                                           retries=retries)

        return self._parse_response(response)

    def _parse_response(self, response):
        """ Try to decode response body to json
        """
        if response.status < 200 or response.status > 300:
            raise InvalidStatusCodeResponse(response.status, response.data)

        if response.headers.get('Content-Type') != 'application/json':
            raise InvalidResponseContentType(response.headers.get('Content-Type'))

        try:
            json_response = json.loads(response.data.decode('utf-8'))
        except ValueError as exc:
            raise InvalidJsonResponse(exc)

        if json_response.get('status', False) is False:
            raise StatusFailedResponse(json_response)

        return json_response



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from setuptools import setup, find_packages
import sqreen

long_description = """Sqreen is a SaaS based Application protection and monitoring platform that integrates directly into your Python applications.
Learn more at `<https://sqreen.io>`_."""


setup(
    name='sqreen',
    version=sqreen.__version__,
    description="Sqreen agent to protect Python applications.",
    long_description=long_description,
    author="Boris Feld",
    author_email='boris@sqreen.io',
    url='https://github.com/sqreen/AgentPython',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_dir={'sqreen':
                 'sqreen'},
    package_data={'sqreen': ['*.crt']},
    include_package_data=True,
    install_requires=['py-mini-racer'],
    license="Copyright",
    zip_safe=False,
    keywords='sqreen',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: Other/Proprietary License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=[],
    entry_points={
        'console_scripts':  [
            "sqreen-start = sqreen.bin.protect:protect"
        ]
    }
)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Flask specific WSGI HTTP Request / Response stuff
"""

from itertools import chain
from logging import getLogger
from traceback import format_stack

try:
    # Python 2
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from .base import BaseRequest


LOGGER = getLogger(__name__)


class FlaskRequest(BaseRequest):

    def __init__(self, request):
        self.request = request

        # Convert django QueryDict to a normal dict with values as list
        self.converted_get = dict(self.request.args.lists())

    @property
    def query_params(self):
        return self.converted_get

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        return list(chain.from_iterable(self.converted_get.values()))

    @property
    def form_params(self):
        return dict(self.request.form)

    @property
    def cookies_params(self):
        return dict(self.request.cookies)

    @property
    def client_ip(self):
        access_route = self.request.access_route
        if len(access_route) == 0:
            return None
        return access_route[0]

    @property
    def hostname(self):
        parsed = urlparse.urlparse(self.request.url_root)
        return parsed.netloc

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return self.request.referrer

    @property
    def client_user_agent(self):
        return self.request.user_agent.string

    @property
    def path(self):
        return self.request.full_path.rstrip('?')

    @property
    def scheme(self):
        return self.request.scheme

    @property
    def server_port(self):
        return self.request.environ.get('SERVER_PORT')

    @property
    def remote_port(self):
        return str(self.request.environ.get('REMOTE_PORT'))

    def get_header(self, name):
        """ Get a specific header
        """
        return self.request.environ.get(name)

    @property
    def caller(self):
        return format_stack()

    @property
    def view_params(self):
        return self.request.view_args



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Batch delivery method
"""
from time import time
from random import randint

from .simple import SimpleDeliverer
from ..remote_exception import RemoteException
from ..events import Attack


class BatchDeliverer(SimpleDeliverer):
    """ Class responsible for batching events before sending them to backend
    """

    def __init__(self, session, max_batch, max_staleness, randomize_staleness=True):
        super(BatchDeliverer, self).__init__(session)
        self.max_batch = max_batch
        self.original_max_staleness = max_staleness
        self.max_staleness = max_staleness
        self.randomize_staleness = randomize_staleness

        self.current_batch = []
        self.first_seen = {}

    def post_event(self, event):
        """ Post a single event
        """
        self.current_batch.append(event)
        if self.post_batch_needed(event):
            self.deliver_batch()

    def post_batch_needed(self, event):
        """ Check if the batch should be posted
        """
        return self._first_seen(event) or self._max_batch_size() or self._stale()

    def _first_seen(self, event):
        """ Check if it's the first time we see an event type, if it's the case
        the batch should be delivered.
        Update the first_seen time
        """
        key = self._event_key(event)
        time_seen = self.first_seen.get(key)

        if time_seen is None:
            self.first_seen[key] = time()

        return time_seen is None

    def deliver_batch(self):
        """ Send the current batch to the backend.
        Reset the first_seen values and randomize the max_staleness
        """
        batch = [self._format_event(event) for event in self.current_batch]
        self.session.post_batch(batch)

        # Clean current state
        self.current_batch = []
        now = time()

        for key in self.first_seen:
            self.first_seen[key] = now

        self._randomize_staleness()

    def _randomize_staleness(self):
        """ Update max_stalenness with up-to 10% of original amount.
        Do nothing if randomize_staleness is False
        """
        if self.randomize_staleness:
            random_decay = randint(0, self.original_max_staleness / 10)
            self.max_staleness = self.original_max_staleness + random_decay

    @staticmethod
    def _format_event(event):
        """ Format an event for the batch depending of its type
        """
        if isinstance(event, RemoteException):
            formatted = event.to_dict()
            formatted['event_type'] = 'sqreen_exception'
        elif isinstance(event, Attack):
            formatted = event.to_dict()
            formatted['event_type'] = 'attack'
        else:
            raise NotImplementedError('Unknown event type {}'.format(type(event)))

        return formatted

    def drain(self):
        """ Drain current batch if not empty
        """
        if self._not_empty_batch():
            self.deliver_batch()

    def tick(self):
        """ Periodic check to send batch when needed
        """
        should_send = self._not_empty_batch() and self._stale()
        if should_send:
            self.deliver_batch()
        return should_send

    @staticmethod
    def _event_key(event):
        """ Unique key for first seen time per event_type
        """
        if isinstance(event, RemoteException):
            return "rex-{}".format(event.exception_class)
        elif isinstance(event, Attack):
            return "att-{}".format(event.rule_name)
        else:
            raise NotImplementedError('Unknown event type {}'.format(type(event)))

    ###
    # Various checkers used for checking if current batch should be sent
    # or not to the backend
    ###

    def _max_batch_size(self):
        """ Check if the batch has reached it max size, if it's the case, the
        batch should be delivered.
        """
        return len(self.current_batch) >= self.max_batch

    def _not_empty_batch(self):
        """ Check if the current batch is not empty.
        """
        return len(self.current_batch) > 0

    def _stale(self):
        """ Check if one event has been seen more than max_staleness in the past
        """
        min_seen = min(self.first_seen.values())
        return (min_seen + self.max_staleness) < time()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Condition Evaluator
"""

import operator

from logging import getLogger

from .binding_accessor import BindingAccessor
from .exceptions import SqreenException
from .utils import is_string


LOGGER = getLogger(__name__)


class ConditionValueError(SqreenException):
    """ Exception raised when the condition is invalid
    """


class ConditionRecursionError(SqreenException):
    """ Exception raised when the condition is too deeply nested
    """


def hash_value_includes(value, iterable, min_value_size, max_iterations=1000):
    """ Tries to find if a string value includes one of the deeply nested
    iterable values. The iterable could be composed of dict / list or a
    combination of the two.
    The min_value_size is used to avoid comparison on small strings. For example
    there is no possible SQL injection below 8 characters.
    """
    iteration = 0

    # Early stop
    if iterable is None:
        return False

    if isinstance(iterable, dict):
        remaining_iterables = list(iterable.values())
    else:
        remaining_iterables = iterable

    while len(remaining_iterables) != 0:

        iteration += 1
        # If we have a very big or nested iterable, returns True and execute the
        # rule
        if iteration >= max_iterations:
            return True

        iterable_value = remaining_iterables.pop(0)

        # If we get an iterable, add it to the list of remaining
        if isinstance(iterable_value, dict):
            remaining_iterables.extend(list(iterable_value.values()))
        elif isinstance(iterable_value, list):
            remaining_iterables.extend(iterable_value)
        else:

            # Check the value
            if not is_string(iterable_value):
                iterable_value = str(iterable_value)

            if len(iterable_value) < min_value_size:
                continue

            if iterable_value in value:
                return True

    return False


def and_(*args):
    """ Return the bool value of and between all values
    """
    return all(args)


def or_(*args):
    """ Return the bool value of or between all values
    """
    return any(args)


OPERATORS = {
    "%and": and_,
    "%or": or_,
    "%equals": operator.eq,
    "%not_equals": operator.ne,
    "%gt": operator.gt,
    "%gte": operator.ge,
    "%lt": operator.lt,
    "%lte": operator.le,
    "%include": operator.contains,
    "%hash_val_include": hash_value_includes,
}

OPERATORS_ARITY = {
    "%equals": 2,
    "%not_equals": 2,
    "%gt": 2,
    "%gte": 2,
    "%lt": 2,
    "%lte": 2,
    "%include": 2,
    "%hash_val_include": 3
}


def is_condition_empty(condition):
    """ Returns True if the condition is no-op

    >>> is_condition_empty(None)
    True
    >>> is_condition_empty(True)
    False
    >>> is_condition_empty(False)
    False
    >>> is_condition_empty({})
    True
    >>> is_condition_empty({"%and": ["true", "true"]})
    False
    """
    if condition is None:
        return True
    elif isinstance(condition, bool):
        return False
    elif isinstance(condition, dict):
        return len(condition) == 0
    else:
        LOGGER.warning("Invalid pre condition type %r", condition)
        return True


def compile_condition(condition, level):
    """ Compile a row condition, replace values by BindingAccessor
    and check operators validity and arity.
    """
    if level <= 0:
        raise ConditionRecursionError("Compile went too deep")

    if isinstance(condition, bool):
        return condition

    compiled = {}

    for _operator, values in condition.items():

        # Check operator validity
        if _operator not in OPERATORS:
            raise ConditionValueError('Unkown operator {}'.format(_operator))

        # Check operator arity
        if len(values) != OPERATORS_ARITY.get(_operator, len(values)):
            msg = 'Bad arity for operator {}: {}'
            raise ConditionValueError(msg.format(_operator, len(values)))

        # Check types
        if not isinstance(values, list):
            msg = "Values should be an array (was {})"
            raise ConditionValueError(msg.format(type(values)))

        compiled_values = []
        for value in values:
            if isinstance(value, bool):
                compiled_values.append(value)
            elif isinstance(value, dict):
                compiled_values.append(compile_condition(value, level - 1))
            elif is_string(value):
                compiled_values.append(BindingAccessor(value))
            else:
                compiled_values.append(BindingAccessor(str(value)))

        compiled[_operator] = compiled_values

    return compiled


def resolve_and_evaluate(value, level, **kwargs):
    """ Take a compiled condition, resolve values and evaluate the result
    """
    resolved = resolve(value, level, **kwargs)
    result = evaluate(resolved)
    return result


def resolve(condition, level, **kwargs):
    """ Takes a condition with BindingAccessor and resolve them
    """
    if level <= 0:
        raise ConditionRecursionError("Resolve went too deep")

    if isinstance(condition, bool):
        return condition

    resolved = {}

    for _operator, values in condition.items():

        resolved_values = []

        for value in values:

            if isinstance(value, bool):
                resolved_values.append(value)
            elif isinstance(value, dict):
                resolved_values.append(resolve_and_evaluate(value, level - 1,
                                                            **kwargs))
            else:
                resolved_values.append(value.resolve(**kwargs))

        resolved[_operator] = resolved_values

    return resolved


def evaluate(resolved_condition):
    """ Evaluate a resolved condition
    """
    if isinstance(resolved_condition, bool):
        return resolved_condition

    elif isinstance(resolved_condition, dict):

        result = True

        # Implicit and between operators
        for operator_name, values in resolved_condition.items():
            operator_callable = OPERATORS[operator_name]

            result = result and operator_callable(*values)

            # Break early
            if result is False:
                return result

        return result
    else:
        msg = "Invalid condition type: {!r}"
        raise ConditionValueError(msg.format(resolved_condition))


class ConditionEvaluator(object):
    """ Evaluate a condition, resolving literals using BindingAccessor.
     {"%and": ["true", "true"]} -> true
     {"%or": ["true", "false"]} -> true
     {"%and": ["false", "true"]} -> false
     {"%equal": ["coucou", "#.args[0]"]} -> "coucou" == args[0]
     {"%hash_val_include": ["toto is a small guy", "#.request_params", 0]} ->
          true if one value of request params in included
          in the sentence 'toto is a small guy'.
    Combine expressions:
     { "%or":
       [
         {"%hash_val_include": ["AAA", "#.request_params", 0]},
         {"%hash_val_include": ["BBB", "#.request_params", 0]},
       ]
     }
    will return true if one of the request_params include either AAA or BBB.
    """

    def __init__(self, condition):
        self.compiled = compile_condition(condition, 10)

    def evaluate(self, **kwargs):
        """ Evaluate the compiled condition and return the results
        """
        return resolve_and_evaluate(self.compiled, level=10, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Base Request class
"""

import logging
from traceback import format_stack

LOGGER = logging.getLogger(__name__)


class BaseRequest(object):

    @property
    def request_payload(self):
        """ Returns current request payload with the backend expected field
        name.
        """
        return {
            'addr': self.client_ip,
            'host': self.hostname,
            'verb': self.method,
            'referer': self.referer,
            'user_agent': self.client_user_agent,
            'path': self.path,
            'scheme': self.scheme,
            'port': self.server_port,
            'rport': self.remote_port,
        }

    @property
    def request_params(self):
        """ Returns all the inputs that are controllable by the user
        """
        return {
            'form': self.form_params,
            'query': self.query_params,
            'cookies': self.cookies_params,
            'other': self.view_params
        }

    def params_contains(self, param):
        """ Return True if the parameter given in input is present in the
        request inputs.
        """
        iteration = 0
        max_iterations = 100

        remaining_iterables = [self.request_params]

        while len(remaining_iterables) != 0:

            iteration += 1
            # If we have a very big or nested iterable, returns False
            if iteration >= max_iterations:
                return False

            iterable_value = remaining_iterables.pop(0)

            # If we get an iterable, add it to the list of remaining
            if isinstance(iterable_value, dict):
                remaining_iterables.extend(list(iterable_value.values()))
            elif isinstance(iterable_value, list):
                remaining_iterables.extend(iterable_value)
            else:
                if iterable_value == param:
                    return True

        return False

    @property
    def full_payload(self):
        """ Return request information in the format accepted by the backend
        """
        return {
            'request': self.request_payload,
            'params': self.request_params
        }

    @property
    def caller(self):
        return format_stack()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Flask hook strategy
"""

import sys

from logging import getLogger

from .base import BaseStrategy
from ..hook_point import FlaskMiddleware
from ..import_hook import ImportHook, get_hook_parent


LOGGER = getLogger(__name__)


def try_trigger_before_first_request_functions_wrapper(original, middleware):

    def wrapped_try_trigger_before_first_request_functions(self, *args, **kwargs):
        LOGGER.debug("Execute try_trigger_before_first_request_functions_wrapper")

        try:
            # Ensure we insert our middleware only once
            if self._got_first_request is False:
                # Insert pre middleware method
                self.before_request_funcs.setdefault(None, []).insert(0, middleware.pre)

                # Insert post middleware method
                self.after_request_funcs.setdefault(None, []).insert(0, middleware.post)
        except Exception:
            LOGGER.warning("Error while inserting our middleware", exc_info=True)

        return original(self, *args, **kwargs)

    return wrapped_try_trigger_before_first_request_functions


class FlaskStrategy(BaseStrategy):
    """ Strategy for Flask peripheric callbacks.

    It injects functions that calls pre and post callbacks in the Flask
    request workflow
    """

    MODULE_NAME = "flask.app"
    HOOK_CLASS = "Flask"
    HOOK_METHOD = "try_trigger_before_first_request_functions"

    def __init__(self, strategy_key, channel, before_hook_point=None):
        super(FlaskStrategy, self).__init__(channel, before_hook_point)
        self.strategy_key = strategy_key

        self.flask_middleware = FlaskMiddleware(self, channel)

        self.hooked = False

    def hook(self, callback):
        """ Accept a callback and store it. If it's the first callback
        for this strategy, actually hook the load_middleware to insert our
        middleware.

        Once hooked, the middleware will call the callbacks at the right moment.
        """
        # Register callback
        self.add_callback(callback)

        # Check if we already hooked at
        if not self.hooked:

            import_hook = ImportHook(self.MODULE_NAME, self.import_hook_callback)
            sys.meta_path.insert(0, import_hook)

            self.hooked = True

    def import_hook_callback(self, module):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module
        """
        hook_parent = get_hook_parent(module, self.HOOK_CLASS)

        original = getattr(hook_parent, self.HOOK_METHOD, None)
        hooked = try_trigger_before_first_request_functions_wrapper(original, self.flask_middleware)
        setattr(hook_parent, self.HOOK_METHOD, hooked)
        LOGGER.debug("Successfully hooked on %s %s", self.MODULE_NAME,
                     self.HOOK_CLASS)

    @classmethod
    def get_strategy_id(cls, callback):
        """ This strategy only hook on
        flask.app::Flask, try_trigger_before_first_request_functions)
        """
        return ("{}::{}".format(cls.MODULE_NAME, cls.HOOK_CLASS), cls.HOOK_METHOD)

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.io/terms.html
""" Deliverer classes package
"""
from .simple import SimpleDeliverer
from .batch import BatchDeliverer

__all__ = ["SimpleDeliverer", "BatchDeliverer"]
